from __future__ import print_function

import platform
import unittest
from copy import deepcopy
from csv import DictReader, DictWriter

if False:
    from fs.tempfs import TempFS
    from rowgenerators.urls import Url


    from old.generators import get_generator
    from rowgenerators import RowGenerator
    from rowgenerators import SourceSpec
    from rowgenerators import parse_url_to_dict
    from rowgenerators import register_proto


    try:
        import fiona
        import shapely
        import pyproj
        geo_installed = True
    except ImportError:
        geo_installed = False


    try:
        import jupyter
        import nbconvert
        # This next import may fail due to version issues, but I'm not sure.
        # It seems to fail on Windows
        from nbconvert.exporters import get_exporter
        jupyter_installed = True
    except ImportError:
        jupyter_installed = False


    try:
        import metatab
        metatab_installed = True
    except ImportError:
        metatab_installed = False

    def data_path(v):
        from os.path import dirname, join
        d = dirname(__file__)
        return join(d, 'test_data', v)


    def script_path(v=None):
        from os.path import dirname, join
        d = dirname(__file__)
        if v is not None:
            return join(d, 'scripts', v)
        else:
            return join(d, 'scripts')


    def sources():
        import csv
        with open(data_path('sources.csv')) as f:
            r = csv.DictReader(f)
            return list(r)


    def cache_fs():
        from fs.tempfs import TempFS

        return TempFS('rowgenerator')


    class BasicTests(unittest.TestCase):
        def compare_dict(self, name, a, b):
            from rowgenerators.util import flatten
            fa = set('{}={}'.format(k, v) for k, v in flatten(a));
            fb = set('{}={}'.format(k, v) for k, v in flatten(b));

            # The declare lines move around a lot, and rarely indicate an error
            fa = {e for e in fa if not e.startswith('declare=')}
            fb = {e for e in fb if not e.startswith('declare=')}

            errors = len(fa - fb) + len(fb - fa)

            if errors:
                print("=== ERRORS for {} ===".format(name))

            if len(fa - fb):
                print("In a but not b")
                for e in sorted(fa - fb):
                    print('    ', e)

            if len(fb - fa):
                print("In b but not a")
                for e in sorted(fb - fa):
                    print('    ', e)

            self.assertEqual(0, errors)

        def test_source_spec_url(self):
            from rowgenerators import SourceSpec, RowGenerator
            from copy import deepcopy

            ss = SourceSpec(url='http://foobar.com/a/b.csv')
            self.assertEqual('b.csv', ss.target_file)
            self.assertIsNone(ss.target_segment)

            ss = SourceSpec(url='http://foobar.com/a/b.zip#a')
            print(ss._url)
            self.assertEqual('a', ss.target_file)
            self.assertIsNone(ss.target_segment)

            ss2 = deepcopy(ss)
            self.assertEqual(ss.target_file, ss2.target_file)
            self.assertIsNone(ss.target_segment)

            ss = SourceSpec(url='http://foobar.com/a/b.zip#a;b')
            self.assertEqual('a', ss.target_file)
            self.assertEqual('b', ss.target_segment)

            ss2 = deepcopy(ss)
            self.assertEqual(ss.target_file, ss2.target_file)
            self.assertEqual(ss.target_segment, ss2.target_segment)

            ss = RowGenerator(
                url='http://public.source.civicknowledge.com/example.com/sources/test_data.zip#renter_cost_excel07.xlsx')
            self.assertEqual('renter_cost_excel07.xlsx', ss.target_file)

            ss2 = deepcopy(ss)
            self.assertEqual(ss.target_file, ss2.target_file)

            for url in (
                    'http://example.com/foo/archive.zip',
                    'http://example.com/foo/archive.zip#file.xlsx',
                    'http://example.com/foo/archive.zip#file.xlsx;0',
                    'socrata+http://example.com/foo/archive.zip'
            ):
                pass

            print(SourceSpec(url='socrata+http://chhs.data.ca.gov/api/views/tthg-z4mf').__dict__)

        def test_run_sources(self):

            cache = cache_fs()

            for sd in sources():
                # Don't have the column map yet.
                if sd['name'] in ('simple_fixed', 'facilities'):
                    continue

                try:

                    ss = SourceSpec(**sd)

                    gen = RowGenerator(cache=cache, **sd)

                    rows = list(gen)

                    self.assertEquals(int(sd['n_rows']), len(rows))
                except Exception as e:
                    print('---')
                    print(sd['name'], e)
                    print(rows[0])
                    print(rows[-1])

        def test_inspect(self):

            from rowgenerators import inspect

            cache = cache_fs()

            us = 'http://www.sandiegocounty.gov/content/dam/sdc/hhsa/programs/phs/CHS/Community%20Profiles/MCH_2010-2013.xlsx'

            ss = SourceSpec(us)

            print(ss.update(target_segment='foo'))
            print(ss._url.rebuild_url(target_segment='foo'))
            return

            for spec in inspect(ss, cache, callback=print):
                print(spec)

        def test_google(self):
            from rowgenerators import SourceSpec, GooglePublicSource
            spec = SourceSpec(url='gs://1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w')

            self.assertEquals('gs', spec.proto)
            self.assertEquals('gs://1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/', spec.url)
            self.assertEquals(
                'https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/export?format=csv',
                GooglePublicSource.download_url(spec))

            self.assertEquals(12004, len(list(spec.get_generator(cache_fs()))))

        def test_zip(self):

            from rowgenerators import enumerate_contents, RowGenerator, SourceError, TextEncodingError

            z = 'http://public.source.civicknowledge.com/example.com/sources/test_data.zip'
            cache = cache_fs()

            for c in enumerate_contents(z, cache):

                print(c.url, c.encoding)

                if c.target_format in ('foo', 'txt'):
                    continue

                gen = RowGenerator(url=c.url)
                try:
                    print(len(list(gen)))
                except (UnicodeDecodeError, TextEncodingError) as e:
                    print("UERROR", c.name, e)
                except SourceError as e:
                    print("ERROR", c.name, e)

        def test_d_and_c(self):
            from csv import DictReader
            from old.fetch import download_and_cache

            from os.path import isfile

            cache = TempFS()

            with open(data_path('sources.csv')) as f:
                for e in DictReader(f):
                    try:
                        d = download_and_cache(SourceSpec(**e), cache)
                    except ModuleNotFoundError:
                        # For when metatab isn't installed.
                        continue

                    self.assertTrue(isfile(d['sys_path']))

        def test_delayed_flo(self):
            from csv import DictReader

            cache = TempFS()
            success = []
            errors = []

            with open(data_path('sources.csv')) as f:
                for e in DictReader(f):

                    if e['name'] in ('simple_fixed',):
                        continue

                    if e['name'] not in ('zip_no_xls',):
                        continue

                    s = SourceSpec(**e)

                    print(s.dict)

                    d = get_generator(s, cache)

                    print(s._url, len(list(d)))

        def test_urls(self):

            headers = "in_url class url resource_url resource_file target_file scheme proto resource_format target_format " \
                      "is_archive encoding target_segment".split()

            import tempfile
            tf = tempfile.NamedTemporaryFile(prefix="rowgen", delete=False)
            temp_name = tf.name
            tf.close()

            # S3 URLS have these fields which need to be removed before writing to CSV files.
            def clean(do):

                for f in ['_orig_url', '_key', '_orig_kwargs', '_bucket_name']:
                    try:
                        del do[f]
                    except KeyError:
                        pass

            with open(data_path('url_classes.csv')) as f, open(temp_name, 'w') as f_out:
                w = None
                r = DictReader(f)
                errors = 0
                for i, d in enumerate(r):

                    url = d['in_url']

                    o = Url(url)

                    do = dict(o.__dict__.items())
                    del do['parts']

                    if w is None:
                        w = DictWriter(f_out, fieldnames=headers)
                        w.writeheader()
                    do['in_url'] = url
                    do['is_archive'] = o.is_archive
                    do['class'] = o.__class__.__name__
                    clean(do)
                    w.writerow(do)

                    d = {k: v if v else None for k, v in d.items()}
                    do = {k: str(v) if v else None for k, v in do.items()}  # str() turns True into 'True'

                    # a is the gague data from url_classes.csv
                    # b is the test object.

                    try:  # A, B
                        self.compare_dict(url, d, do)
                    except AssertionError as e:
                        errors += 1
                        print(e)
                        # raise

                self.assertEqual(0, errors)

            with open(data_path('url_classes.csv')) as f:

                r = DictReader(f)
                for i, d in enumerate(r):
                    u1 = Url(d['in_url'])

            with open(data_path('url_classes.csv')) as f:

                r = DictReader(f)
                for i, d in enumerate(r):
                    u1 = Url(d['in_url'])
                    d1 = u1.__dict__.copy()
                    d2 = deepcopy(u1).__dict__.copy()

                    # The parts will be different Bunch objects
                    clean(d1)
                    clean(d2)
                    del d1['parts']
                    del d2['parts']

                    self.assertEqual(d1, d2)

                    self.assertEqual(d1, u1.dict)

            for us in ("http://example.com/foo.zip", "http://example.com/foo.zip#a;b"):
                u = Url(us, encoding='utf-8')
                u2 = u.update(target_file='bingo.xls', target_segment='1')

                self.assertEqual('utf-8', u2.dict['encoding'])
                self.assertEqual('bingo.xls', u2.dict['target_file'])
                self.assertEqual('1', u2.dict['target_segment'])

        def test_url_update(self):

            u1 = Url('http://example.com/foo.zip')

            self.assertEqual('http://example.com/foo.zip#bar.xls', u1.rebuild_url(target_file='bar.xls'))
            self.assertEqual('http://example.com/foo.zip#0', u1.rebuild_url(target_segment=0))
            self.assertEqual('http://example.com/foo.zip#bar.xls%3B0',
                             u1.rebuild_url(target_file='bar.xls', target_segment=0))

            u2 = u1.update(target_file='bar.xls')

            self.assertEqual('bar.xls', u2.target_file)
            self.assertEqual('xls', u2.target_format)

            self.assertEqual('http://example.com/foo.zip', u1.rebuild_url(False, False))

            self.assertEqual('file:metatadata.csv', Url('file:metatadata.csv').rebuild_url())

        def test_parse_file_urls(self):
            from rowgenerators.util import parse_url_to_dict, unparse_url_dict
            urls = [
                ('file:foo/bar/baz', 'foo/bar/baz', 'file:foo/bar/baz'),
                ('file:/foo/bar/baz', '/foo/bar/baz', 'file:/foo/bar/baz'),
                ('file://example.com/foo/bar/baz', '/foo/bar/baz', 'file://example.com/foo/bar/baz'),
                ('file:///foo/bar/baz', '/foo/bar/baz', 'file:/foo/bar/baz'),
            ]

            for i, o, u in urls:
                p = parse_url_to_dict(i)
                self.assertEqual(o, p['path'])
                self.assertEqual(u, unparse_url_dict(p))
                # self.assertEqual(o, parse_url_to_dict(u)['path'])



        def test_metatab_url(self):

            urlstr = 'metatab+http://s3.amazonaws.com/library.metatab.org/cdss.ca.gov-residential_care_facilities-2017-ca-7.csv#facilities'

            u = Url(urlstr)

            self.assertEqual('http', u.scheme)
            self.assertEqual('metatab', u.proto)
            self.assertEqual('http://s3.amazonaws.com/library.metatab.org/cdss.ca.gov-residential_care_facilities-2017-ca-7.csv', u.resource_url)
            self.assertEqual('cdss.ca.gov-residential_care_facilities-2017-ca-7.csv', u.target_file)
            self.assertEqual('facilities', u.target_segment)

        @unittest.skipIf(not metatab_installed, "Metatab modules are not installed")
        def test_metapack(self):

            from metatab import open_package, resolve_package_metadata_url

            cache = cache_fs()

            url = 'metatab+http://library.metatab.org/example.com-simple_example-2017-us-1#random-names'

            rg = RowGenerator(cache=cache, url=url)

            package_url, metadata_url = resolve_package_metadata_url(rg.generator.spec.resource_url)

            self.assertEquals('http://library.metatab.org/example.com-simple_example-2017-us-1/',package_url)
            self.assertEquals('http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv',metadata_url)

            doc = open_package(rg.generator.spec.resource_url, cache=cache)

            self.assertEquals('http://library.metatab.org/example.com-simple_example-2017-us-1/data/random-names.csv',
                              doc.resource('random-names').resolved_url)


            urls = ['metatab+http://library.metatab.org/example.com-simple_example-2017-us-1#random-names',
                    'metatab+http://library.metatab.org/example.com-simple_example-2017-us-1.zip#random-names',
                    'metatab+http://library.metatab.org/example.com-simple_example-2017-us-1.xlsx#random-names'
                    ]

            for url in urls:
                gen = None
                try:
                    gen = RowGenerator(cache=cache, url=url)

                    rows = list(gen)

                    self.assertEquals(101, len(rows))
                except:
                    print("ERROR URL", url)
                    print("Row Generator ", gen)
                    raise

        @unittest.skipIf(platform.system() == 'Windows','ProgramSources don\'t work on Windows')
        def test_program(self):

            urls = (
                ('program:rowgen.py', 'rowgen.py'),
                ('program:/rowgen.py', '/rowgen.py'),
                ('program:///rowgen.py', '/rowgen.py'),
                ('program:/a/b/c/rowgen.py', '/a/b/c/rowgen.py'),
                ('program:/a/b/c/rowgen.py', '/a/b/c/rowgen.py'),
                ('program:a/b/c/rowgen.py', 'a/b/c/rowgen.py'),
                ('program+http://foobar.com/a/b/c/rowgen.py', '/a/b/c/rowgen.py'),
            )

            for u, v in urls:
                url = Url(u)

                self.assertEquals(url.path, v, u)

            cache = cache_fs()

            options = {
                '-a': 'a',
                '-b': 'b',
                '--foo': 'foo',
                '--bar': 'bar'
            }

            options.update({'ENV1': 'env1', 'ENV2': 'env2', 'prop1': 'prop1', 'prop2': 'prop2'})

            gen = RowGenerator(cache=cache, url='program:rowgen.py', working_dir=script_path(),
                               generator_args=options)

            rows = list(gen)

            for row in rows:
                print(row)

        @unittest.skipIf(not jupyter_installed, "Juptyer notebook modules are not installed")
        def test_notebook(self):

            urls = (
                'ipynb+file:foobar.ipynb',
                'ipynb+http://example.com/foobar.ipynb',
                'ipynb:foobar.ipynb'

            )

            for url in urls:
                u = Url(url)
                print(u, u.path, u.resource_url)

                s = SourceSpec(url)
                print(s, s.proto, s.scheme, s.resource_url, s.target_file, s.target_format)
                self.assertIn(s.scheme, ('file', 'http'))
                self.assertEquals('ipynb', s.proto)
                # print(download_and_cache(s, cache_fs()))

            gen = RowGenerator(cache=cache_fs(),
                               url='ipynb:Py3Notebook.ipynb#lst',
                               working_dir=script_path(),
                               generator_args={'mult': lambda x: x * 3})

            print(gen.generator.execute())

        @unittest.skipIf(not geo_installed,"geo modules are not installed")
        def test_shapefile(self):

            url = "shape+http://s3.amazonaws.com/test.library.civicknowledge.com/census/tl_2016_us_state.geojson"

            gen = RowGenerator(url=url, cache=cache_fs())

            self.assertTrue(gen.is_geo)

            print("HEADERS", gen.headers)

            x = 0
            for row in gen.iter_rp():
                x += float(row['INTPTLON'])

            self.assertEquals(-4776, int(x))

            url = "shape+http://s3.amazonaws.com/test.library.civicknowledge.com/census/tl_2016_us_state.geojson.zip"

            gen = RowGenerator(url=url, cache=cache_fs())

            self.assertTrue(gen.is_geo)

            x = 0
            for row in gen.iter_rp():
                x += float(row['INTPTLON'])

            self.assertEquals(-4776, int(x))

            return

            url = "shape+http://s3.amazonaws.com/test.library.civicknowledge.com/census/tl_2016_us_state.zip"

            gen = RowGenerator(url=url, cache=cache_fs())

            for row in gen:
                print(row)




        def test_windows_urls(self):

            url = 'w:/metatab36/metatab-py/metatab/templates/metatab.csv'

            print(parse_url_to_dict(url))

            url = 'N:/Desktop/metadata.csv#renter_cost'

            print(parse_url_to_dict(url))

        def test_query_urls(self):

            url='https://s3.amazonaws.com/private.library.civicknowledge.com/civicknowledge.com-rcfe_health-1/metadata.csv?AWSAccessKeyId=AKIAJFW23EPQCLXRU7DA&Signature=A39XhRP%2FTKAxv%2B%2F5vCubwWPDag0%3D&Expires=1494223447'

            u = Url(url)

            print(u.resource_file, u.resource_format)
            print(u.target_file, u.target_format)


        def test_s3_url(self):

            from rowgenerators.urls import S3Url

            url_str = 's3://bucket/a/b/c/file.csv'

            u = Url(url_str)

            self.assertEquals(S3Url, type(u))


        def test_register(self):

            from pandasreporter import CensusReporterSource, get_cache

            register_proto('censusreporter', CensusReporterSource)

            url = 'censusreporter:B17001/140/05000US06073'

            gen = RowGenerator(cache=get_cache(), url=url)

            self.assertEquals('B17001', gen.generator.table_id )
            self.assertEquals('140', gen.generator.summary_level)
            self.assertEquals('05000US06073', gen.generator.geoid)

            for row in gen:
                print(row)


    if __name__ == '__main__':
        unittest.main()



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of the
Revised BSD License, included in this distribution as LICENSE.txt
"""

from .appurl.util import clean_cache, nuke_cache, get_cache

# Keep code cleanups from deleteing these they are imported here
# because they got moved to appurl
assert clean_cache
assert get_cache
assert nuke_cache


def real_files_in_zf(zf):
    """Return a list of internal paths of real files in a zip file, based on the 'external_attr' values"""
    from os.path import basename

    for e in zf.infolist():

        if basename(e.filename).startswith('__') or basename(e.filename).startswith('.'):
            continue

        # I really don't understand external_attr, but no one else seems to either,
        # so we're just hacking here.
        # e.external_attr>>31&1 works when the archive has external attrs set, and a dir heirarchy
        # e.external_attr==0 works in cases where there are no external attrs set
        # e.external_attr==32 is true for some single-file archives.
        if bool(e.external_attr >> 31 & 1 or e.external_attr == 0 or e.external_attr == 32):
            yield e.filename


def fs_join(*args):
    """Like os.path.join, but never returns '\' chars"""
    from os.path import join

    return join(*args).replace('\\','/')



def qualified_class_name(o):
    """Full name of an object, including the module"""
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + '.' + o.__class__.__name__

# From http://stackoverflow.com/a/295466
def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.type(
    """
    import re
    import unicodedata

    value = str(value)
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('utf8').strip().lower()
    value = re.sub(r'[^\w\s\-\.]', '', value)
    value = re.sub(r'[-\s]+', '-', value)
    return value

def memoize(obj):
    from functools import wraps

    cache = obj.cache = {}

    @wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]

    return memoizer

def md5_file(path_or_fh):
    """Hash a large file in chunks"""
    import hashlib

    def hash_fh(fh):
        m = hashlib.md5()
        while True:
            data = fh.read(8192)
            if not data:
                break
            m.update(data)
        return m.hexdigest()

    try:
        return hash_fh(path_or_fh)
    except AttributeError:
        with open(path_or_fh, 'rb') as fh:
            return hash_fh(fh)


class Constant:
    """Organizes constants in a class."""

    class ConstError(TypeError):
        pass

    def __setattr__(self, name, value):
        if name in self.__dict__:
            raise self.ConstError("Can't rebind const(%s)" % name)
        self.__dict__[name] = value



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """


class DictRowGenerator(object):
    """Constructed on a RowGenerator, returns dicts from the second and subsequent rows, using the
    first row as dict keys. """

    def __init__(self, rg):
        self._rg = rg

    def __iter__(self):


        headers = None

        for row in self._rg:
            if not headers:
                headers = [str(e).strip() for e in row]
                continue

            yield dict(zip(headers, row))




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from rowgenerators.appurl.web.web import WebUrl
from rowgenerators.appurl.url import parse_app_url
from rowgenerators.appurl.file.csv import CsvFileUrl

class GoogleProtoCsvUrl(WebUrl):
    """Access a Google spreadheet as a CSV format download"""

    csv_url_template = 'https://docs.google.com/spreadsheets/d/{key}/export?format=csv'

    def __init__(self, url, **kwargs):
        kwargs['resource_format'] = 'csv'
        kwargs['encoding'] = 'utf8'
        kwargs['proto'] = 'gs'
        super(GoogleProtoCsvUrl, self).__init__(url, **kwargs)

    @classmethod
    def match(cls, url, **kwargs):
        return url.proto == 'gs'

class GoogleSpreadsheetUrl(WebUrl):

    """A Google spreadsheet url. Supports URLs with these forms:

        gs://1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w
        gs+https://drive.google.com/open?id=1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w
        gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing
        gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing
        gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit#gid=801701031

        You can also leave off the 'gs+', so these are also valid:

        https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing
        https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing
        https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit#gid=801701031

        """

    match_priority = WebUrl.match_priority - 1

    url_template = 'https://docs.google.com/spreadsheets/d/{key}/export?format=csv'
    gid_siffix = '&gid={gid}'

    def __init__(self, url=None, downloader=None, **kwargs):

        super().__init__(url, downloader, **kwargs)

        parts = self.path.split('/') + [self.query.replace('id=', ''), self.netloc]
        parts = list(reversed(sorted(parts, key=lambda k: len(k))))
        self.key = parts[0]

        if self.scheme == 'gs':
            self.gid = self.target_file
        else:
            self.gid = self.fragment_query.get('gid')

        if self.gid:
            web_url = (self.url_template + self.gid_siffix).format(key=self.key, gid=self.gid)
        else:
            web_url = self.url_template .format(key=self.key)

        web_url +="#target_file={}-{}.csv".format(self.key,self.gid)

        self.web_url = WebUrl(web_url, downloader=self.downloader)

    @classmethod
    def _match(cls, url, **kwargs):

        return url.proto == 'gs' or \
               (url.netloc == 'docs.google.com' and url.path.startswith('/spreadsheets'))

    def get_resource(self):

        return CsvFileUrl(str(self.web_url.get_resource()))

    def get_target(self):
        return self.get_resource().get_target()








from . import DEFAULT_EPSG

def geoframe(url):
    """Parse an App url and return a Geopandas geoframe"""
    pass



def dataframe(url,  downloader='default', *args, **kwargs):
    """Parse an App url and return a Pandas dataframe"""
    from . import parse_app_url
    import pandas as pd
    from itertools import islice

    u = parse_app_url(url, downloader=downloader, **kwargs)
    r = u.get_resource()
    t = r.get_target()
    g = t.generator

    # Try the generator's dataframe method, if it has one
    try:
        return g.dataframe( *args, **kwargs)
    except AttributeError:
        pass

    # Just normal data, so use the iterator in this object.
    headers = next(islice(g, 0, 1))
    data = islice(g, 1, None)

    df = pd.DataFrame(list(data), columns=headers,  *args, **kwargs)

    df.metatab_errors = g.errors if hasattr(g, 'errors') and g.errors else {}

    return df

def geoframe(url, downloader='default', *args, **kwargs):
    """Return a Geo dataframe"""

    from rowgenerators import parse_app_url
    from .exceptions import SourceError
    from geopandas import GeoDataFrame
    import geopandas as gpd
    from shapely.geometry.polygon import BaseGeometry
    from shapely.wkt import loads

    u = parse_app_url(url, downloader=downloader, **kwargs)

    # Some base URLs have the geoframe.
    try:
        return u.geoframe(*args, **kwargs)
    except AttributeError:
        pass


    r = u.get_resource()
    t = r.get_target()


    g = t.generator

    # Try the generator
    try:
        return g.geoframe(*args, **kwargs)
    except AttributeError:
        pass

    # Maybe the target has a geoframe
    try:
        return t.geoframe(*args, **kwargs)
    except AttributeError:
        pass



    try:

        gdf = GeoDataFrame(dataframe(url, *args, **kwargs))

        first = next(gdf.iterrows())[1]['geometry']

        if isinstance(first, str):
            # We have a GeoDataframe, but the geometry column is still strings, so
            # it must be converted
            shapes = [loads(row['geometry']) for i, row in gdf.iterrows()]

        elif not isinstance(first, BaseGeometry):
            # If we are reading a metatab package, the geometry column's type should be
            # 'geometry' which will give the geometry values class type of
            # rowpipe.valuetype.geo.ShapeValue. However, there are other
            # types of objects that have a 'shape' property.

            shapes = [row['geometry'].shape for i, row in gdf.iterrows()]

        else:
            shapes = gdf['geometry']

        gdf['geometry'] = gpd.GeoSeries(shapes)
        gdf.set_geometry('geometry')

        # Wild guess. This case should be most often for Metatab processed geo files,
        # which are all 4326
        if gdf.crs is None:
            gdf.crs = {'init': f'epsg:{DEFAULT_EPSG}'}

    except KeyError as e:
        raise SourceError("Failed to create GeoDataFrame for resource '{}': No geometry column".format(t))
    except (KeyError,TypeError) as e:
        raise SourceError("Failed to create GeoDataFrame for resource '{}': {}".format(t, str(e)))

    return gdf


def generator(url, downloader='default', *args, **kwargs):
    """Parse an App and return a generator"""
    from rowgenerators import parse_app_url

    u = parse_app_url(url, downloader=downloader, **kwargs)
    r = u.get_resource()
    t = r.get_target()
    return  t.generator


def iterator(url, downloader='default', *args, **kwargs):
    """Parse an App and return a row iterator"""

    from rowgenerators import parse_app_url

    u = parse_app_url(url, downloader=downloader, **kwargs)
    r = u.get_resource()
    t = r.get_target()
    g = t.generator

    return iter(g)



def get_generator(source,  **kwargs):
    """ Locate a generator from the entrypoints.

    """
    import inspect
    import collections
    from pkg_resources import iter_entry_points
    from rowgenerators.exceptions import RowGeneratorError
    from rowgenerators.appurl import parse_app_url, Url
    from rowgenerators.source import Source

    names = []

    if isinstance(source, Source):
        return source

    if isinstance(source, str):

        ref = parse_app_url(source).get_resource().get_target()
        try:
            names.append('.{}'.format(ref.target_format))
        except AttributeError:
            pass

    elif inspect.isgenerator(source):
        names.append('<generator>')
        ref = source

    elif isinstance(source, collections.abc.Iterable):
        names.append('<iterator>')
        ref = source

    elif hasattr(source, '__iter__'):
        names.append('<iterator>')
        ref = source

    elif isinstance(source, Url):

        # Create all of the possible names that this URL could match with

        ref = source
        try:
            names.append('.{}'.format(ref.target_format))
        except AttributeError:
            pass

        try:
            names.append('{}+'.format(ref.scheme_extension))
        except AttributeError:
            pass

        try:
            names.append('{}:'.format(ref.scheme))
        except AttributeError:
            pass

        try:
            names.append('<{}>'.format(ref.__class__.__name__))
        except AttributeError:
            pass

        try:
            if ref.generator_class:
                names.append('<{}>'.format(ref.generator_class.__name__))
        except AttributeError:
            pass

        try:
            names.append('{}+.{}'.format(ref.scheme_extension, ref.target_format))
        except AttributeError:
            pass


    else:
        raise RowGeneratorError("Unknown arg type for source {}, type='{}'".format(source, type(source)))

    classes = sorted([ep.load() for ep in iter_entry_points(group='rowgenerators') if ep.name in names],
                     key=lambda cls: cls.priority)

    if not classes:
        raise RowGeneratorError(("Can't find generator for url '{}' \ntype={}, proto={}, "
                                  "resource_format={}, target_format={}, names={} ")
                                 .format( source, type(source), ref.proto, ref.resource_format, ref.target_format, names))
    try:
        return classes[0](ref, **kwargs)
    except NotImplementedError:
        raise
    except Exception as e:
        raise RowGeneratorError("Failed to instantiate generator for class '{}', ref '{}'".format(classes[0],
                                                                                                   ref)) from e

class SelectiveRowGenerator(object):
    """Proxies an iterator to remove headers, comments, blank lines from the row stream.
    The header will be emitted first, and comments are avilable from properties """

    def __init__(self, seq, start=0, headers=[], comments=[], end=[], load_headers=True, **kwargs):
        """
        An iteratable wrapper that coalesces headers and skips comments

        :param seq: An iterable
        :param start: The start of data row
        :param headers: An array of row numbers that should be coalesced into the header line, which is yieled first
        :param comments: An array of comment row numbers
        :param end: The last row number for data
        :param kwargs: Ignored. Sucks up extra parameters.
        :return:
        """

        self.iter = iter(seq)
        self.start = start if (start or start == 0) else 1
        self.header_lines = headers if isinstance(headers, (tuple, list)) else [int(e) for e in headers.split(',') if e]
        self.comment_lines = comments
        self.end = end

        self.load_headers = load_headers

        self.headers = []
        self.comments = []

        int(self.start)  # Throw error if it is not an int

    @property
    def coalesce_headers(self):
        """Collects headers that are spread across multiple lines into a single row"""

        import re

        if not self.headers:
            return None

        header_lines = [list(hl) for hl in self.headers if bool(hl)]

        if len(header_lines) == 0:
            return []

        if len(header_lines) == 1:
            return header_lines[0]

        # If there are gaps in the values of a line, copy them forward, so there
        # is some value in every position
        for hl in header_lines:
            last = None
            for i in range(len(hl)):
                hli = str(hl[i])
                if not hli.strip():
                    hl[i] = last
                else:
                    last = hli

        headers = [' '.join(str(col_val).strip() if col_val else '' for col_val in col_set)
                   for col_set in zip(*header_lines)]

        headers = [re.sub(r'\s+', ' ', h.strip()) for h in headers]

        return headers

    def __iter__(self):

        for i, row in enumerate(self.iter):

            if i in self.header_lines:
                if self.load_headers:
                    self.headers.append(row)
            elif i in self.comment_lines:
                self.comments.append(row)
            elif i == self.start:
                break

        if self.headers:

            headers = self.coalesce_headers
        else:
            headers = ['col' + str(i) for i, _ in enumerate(row)]

        yield headers

        yield row

        yield from self.iter



# -*- coding: utf-8 -*-

DEFAULT_EPSG=4326

from .appurl.url import parse_app_url, Url
from .appurl.util import get_cache, set_default_cache_name
from .core import get_generator, dataframe, geoframe, iterator
from .source import  Source, RowGenerator
from .appurl.web.download import Downloader
from .exceptions import SourceError


from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass





# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """


from functools import partial
from itertools import islice

from rowgenerators.appurl.file.shapefile import ShapefileUrl, ShapefileShpUrl
from rowgenerators.source import Source
from rowgenerators.exceptions import RowGeneratorError

default_crs = {'init': 'epsg:4326'}

# Looks like PyPy doesn't have ModuleNotFoundError

try:
    ModuleNotFoundError
except NameError:
    class ModuleNotFoundError(ImportError):
        pass


def _import_requirements():
    """Check that the requirements are available. These are in a seperate function because test collectors will
    import the file looking for tests, causing a failure, even if no geo tests are run"""
    try:
        import fiona
        from fiona.crs import from_epsg
        from shapely.geometry import asShape
        from shapely.ops import transform
        import pyproj
    except (ModuleNotFoundError, ImportError) as e:
        raise ImportError("Using ShapefileSource requires installing fiona, shapely and pyproj ") from e
        pass  # HACK Because this file gets collected by the test collectors

class GeoSourceBase(Source):
    """ Base class for all geo sources. """
    pass


    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        """
        A row source for shapefiles.

        By default will try to re-project to epsg:4326 during iteration. This can be turned off by setting
        the 'projection' argument to '<source>'

        :param ref:
        :param cache:
        :param working_dir:
        :param env:
        :param projection: Either an EPSG string, defaults to 'epsg:4326', or '<source>' to not project
        :param kwargs:
        """

        super().__init__(ref, cache, working_dir)

        _import_requirements()

        self.property_schema = self._parameters

        self._kwargs = kwargs

        target_projection = (env or {}).get('projection', default_crs['init']).lower()

        try:
            int(target_projection)
            target_projection = 'epsg:{}'.format(target_projection)
        except ValueError:
            if not target_projection.startswith('epsg:') and target_projection != '<source>':
                raise RowGeneratorError("ShapefileSource projection property must start with 'epsg:'  or be an integer")

        self.target_projection = target_projection
        self.source_projection = None

        # Holds metadata, such as EPSG, that is inferred during processing.
        self._meta = {}


    @property
    def meta(self):
        '''Used as post-iterator metadata in Metapack, copied into a resource term after build'''
        return self._meta

    @property
    def headers(self):
        """Return headers. This must be run after iteration, since the value that is returned is
        set in iteration """

        # self.spec.columns = [c for c in self._get_columns(property_schema)]

        return [x['name'] for x in self.columns]

    @property
    def columns(self):
        """ Returns columns for the file accessed by accessor.

        """
        #
        # first column is id and will contain id of the shape.
        columns = [{'name': 'id', 'type': 'int'}]

        # extend with *.shp file columns converted to ambry_sources format.
        columns.extend(list(map(self._convert_column, self.property_schema.items())))

        # last column is wkt value.
        columns.append({'name': 'geometry', 'type': 'geometry_type'})
        return columns

    def _convert_column(self, shapefile_column):
        """ Converts column from a *.shp file to the column expected by ambry_sources."""
        name, type_ = shapefile_column
        type_ = type_.split(':')[0]
        return {'name': name, 'type': type_}

    def __iter__(self):
        """ Returns generator over shapefile rows.

        Note:
            The first column is an id field, taken from the id value of each shape
            The middle values are taken from the property_schema
            The last column is a string named geometry, which has the wkt value, the type is geometry_type.

        """

        # These imports are nere, not at the module level, so the geo
        # support can be an extra

        import fiona
        from fiona.crs import from_epsg
        from shapely.geometry import shape
        from shapely.ops import transform
        import pyproj

        self.start()

        args, kwargs = self._fiona_open_params()

        with fiona.open(*args, **kwargs ) as source:

            if self.target_projection == '<source>':
                self.target_projection = source.crs.get('init')

            self.source_projection = source.crs.get('init')

            if self.source_projection != self.target_projection:

                projection_type, self.epsg_propjection_code = self.target_projection.split(':')

                assert projection_type == 'epsg'

                int(self.epsg_propjection_code)

                transformer = pyproj.Transformer.from_crs(self.source_projection,
                                                          f'epsg:{self.epsg_propjection_code}')

                def project(x,y,z=None):
                    return transformer.transform(x,y)

                self.projection = self.target_projection
            else:
                project = None
                self.projection = self.source_projection

            self._meta['source_projection'] = self.source_projection
            self._meta['target_projection'] = self.target_projection
            self._meta['projection'] = self.projection

            yield self.headers

            for i,s in enumerate(source):

                row_data = s['properties']

                try:
                    shp = shape(s['geometry'])
                except AttributeError as e:
                    shp = None

                row = [int(s['id'])]
                for col_name, elem in row_data.items():
                    row.append(elem)


                if project and shp:
                    row.append(transform(project, shp))

                else:
                    row.append(shp)

                yield row

        self.finish()


class ShapefileSource(GeoSourceBase):
    """ Accessor for shapefiles (*.shp) with geo data. """

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):

        assert isinstance(ref, (ShapefileUrl, ShapefileShpUrl))

        super().__init__(ref, cache, working_dir, env, **kwargs)


    def _open_file_params(self):
        from zipfile import ZipFile

        layer_index = self.ref.target_segment or 0

        if self.ref.resource_format == 'zip':
            # Find the SHP file. I thought Fiona used to do this itself ...
            assert self.ref.target_file

            vfs = 'zip://{}'.format(self.ref.fspath)

            if self.ref.target_file:
                shp_file = '/' + self.ref.target_file.strip('/')
            else:
                shp_file = '/' + next(
                    n for n in ZipFile(self.ref.fspath).namelist() if (n.endswith('.shp') or n.endswith('geojson')))
        else:
            shp_file = self.ref.fspath
            vfs = None

        return vfs, shp_file, layer_index

    @property
    def _parameters(self):
        import fiona

        vfs, shp_file, layer_index = self._open_file_params()

        with fiona.open(vfs, layer=layer_index) as source:

            return source.schema['properties']

    def _fiona_open_params(self):
        vfs, shp_file, layer_index = self._open_file_params()

        return (vfs, ), dict( layer=layer_index)


    def dataframe(self, limit=None, *args, **kwargs):
        """An alias for geoframe()"""

        return self.geoframe()


    def geoframe(self, *args, **kwargs):
        """Return a geopandas dataframe. The geoframe does not reproject, ( which is a lot faster )
        but does set the crs with the actual projection, so you can re-project with to_crs()

        """
        import geopandas as gpd

        vfs, shp_file, layer_index = self._open_file_params()

        o = gpd.read_file(vfs, layer=layer_index)

        return o

class GeoJsonSource(GeoSourceBase):
    """Generate rows, of Shapley objects, from a GeoJson file reference"""

    delimiter = ','

    @property
    def _parameters(self):
        import fiona

        t = self.ref.get_resource().get_target()

        with fiona.open(t.fspath) as source:
            return source.schema['properties']

    def _fiona_open_params(self):

        t = self.ref.get_resource().get_target()

        return (t.fspath,), {}




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""Base class for Web URLs. These are URLs that can be fetched to the local filesystem. """


from rowgenerators.appurl.util import parse_url_to_dict
from rowgenerators.appurl.url import Url

class WebUrl(Url):
    """Base class for web urls.

    This documentation only describes the differences in implementation from the super class.
    See the documentation for the superclass, :py:class:`appurl.Url` for the default implementations.

    """

    match_priority = 20

    def __init__(self, url=None, downloader=None, **kwargs):

        self._resource = None  # return value from the downloader

        super().__init__(url,downloader=downloader, **kwargs)



    @classmethod
    def _match(cls, url, **kwargs):
        """Return True if this handler can handle the input URL"""
        return url.scheme.startswith('http')

    def list(self):
        """Return a list of this URL with the fragments from listing the resource"""

        r = self.get_resource()

        return list(self.set_target_file(u.target_file) for u in r.list())


    @property
    def auth_resource_url(self):
        """Return An ``S3:`` version of the url, with a resource_url format that will trigger boto auth"""

        # This is just assuming that the url was created as a resource from the S2Url, and
        # has the form 'https://s3.amazonaws.com/{bucket}/{key}'

        parts = parse_url_to_dict(self.resource_url)

        return 's3://{}'.format(parts['path'])

    def get_resource(self):
        """Get the contents of resource and save it to the cache, returning a file-like object"""
        from rowgenerators import parse_app_url # Here, to break an import cycle

        self._resource = self._downloader.download(self.inner)


        ru = parse_app_url(self._resource.sys_path,
                           downloader=self.downloader,
                           scheme_extension=self.scheme_extension,
                           **self.frag_dict)


        return ru

    def dirname(self):
        from os.path import dirname
        return dirname(self.path)

    def basename(self):
        from os.path import basename
        return basename(self.path)

    def join_dir(self, s):


        if self.resource_format in ('zip','xlsx'):
            u = Url(s)
            return self.clone(fragment=u.path)
        else:
            return super().join_dir(s)

    def join_target(self, tf):

            try:
                tf = str(tf.path)
            except:
                pass

            if self.target_format:
                u = self.clone()
                u.target_file = tf

            else:
                # Assuming that if there is no target format, there is no actual target file
                # and the URL is specifying a directory.
                u = self.join(tf)

            return u

class FtpUrl(WebUrl):

    @classmethod
    def _match(cls, url, **kwargs):
        """Return True if this handler can handle the input URL"""
        return url.scheme.startswith('ftp')





from __future__ import print_function

import unittest

from rowgenerators.rowpipe.json import add_to_struct, VTEncoder

class TestJson(unittest.TestCase):


    def test_basic(self):
        import json

        d = {}

        add_to_struct(d, 'a', 1)
        add_to_struct(d, 'b', 2)
        add_to_struct(d, 'c.a', 1)
        add_to_struct(d, 'c.b', 1)
        add_to_struct(d, 'd[]', 1)
        add_to_struct(d, 'd[]', 2)
        add_to_struct(d, 'e[].a', 10)
        add_to_struct(d, 'e[].b', 11)
        add_to_struct(d, 'f[].a[]', 20)
        add_to_struct(d, 'f[].a[]', 21)
        add_to_struct(d, 'f[].b[]', 30)
        add_to_struct(d, 'f[].b[]', 31)
        add_to_struct(d, 'attr[].key', 'k1')
        add_to_struct(d, 'attr[-].value', 'v1')
        add_to_struct(d, 'attr[].key', 'k2')
        add_to_struct(d, 'attr[-].value', 'v2')

        print (json.dumps(d, indent=4))

    @unittest.skip('Has loacal path')
    def test_table(self):
        import json
        from metapack import open_package
        from itertools import islice

        u = '/Volumes/Storage/proj/virt/data-projects/client-boston-college/bc.edu-dataconv_poc/_packages/bc.edu-dataconv_poc-1/'
        pkg = open_package(u)
        r = pkg.resource('comments')

        json_headers = [ (c['pos'], c.get('json')) for c in r.columns()]

        for row in islice(r, None, 10):
            d = {}
            for pos, jh in json_headers:
                add_to_struct(d, jh, row[pos])

            print(json.dumps(d, indent=4, cls=VTEncoder))

if __name__ == '__main__':
    unittest.main()




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from .file import FileUrl

class Hdf5Url(FileUrl):
    """URL that references an HDF5 file"""

    match_priority = FileUrl.match_priority-5

    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        return url.resource_format in ('h5', 'hdf5')

    @property
    def target_format(self):
        return 'h5'

    def list(self, list_self=False):
        import h5py

        with h5py.File(str(self.fspath)) as f:
            return list(f.keys())

    @property
    def generator(self):
        """
        Return the generator for this URL, if the rowgenerator package is installed.

        :return: A row generator object.
        """

        from rowgenerators.core import get_generator

        return get_generator(self, source_url=self)



"""

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

from .core import *
import re



class RaceEthVT(IntDimension):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth'
    lom = LOM.NOMINAL

    # Civick Numeric, Civick, Census, HCI Name, HCI Code, Description
    re_codes = [
        [1, u'aian',        u'C', u'AIAN',      1,    u'American Indian and Alaska Native Alone'],
        [2, u'asian',       u'D', u'Asian',     2,    u'Asian Alone'],
        [3, u'black',       u'B', u'AfricanAm', 3,    u'Black or African American Alone'],
        [4, u'hisp',        u'I', u'Latino',    4,    u'Hispanic or Latino'],
        [5, u'nhopi',       u'E', u'NHOPI',     5,    u'Native Hawaiian and Other Pacific Islander Alone'],
        [6, u'white',       u'A', u'White',     6,    u'White alone'],
        [61, u'whitenh',    u'H', None,         None, u'White Alone, Not Hispanic or Latino'],
        [7, u'other',       None, u'Other',     7,    u'Some Other Race Alone'],
        [8, u'multiple',    None, u'Multiple',  8,    u'Multiple'],
        [9, u'all',         None, u'Total',     9,    u'Total Population']]

    civick_map = {e[0]:e for e in re_codes}

    def __init__(self, v):
        pass

class RaceEthCodeHCI(IntDimension):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/hci'
    lom = LOM.ORDINAL

    hci_map = {e[4]: e[0] for e in RaceEthVT.re_codes if e[4] is not None}


    @property
    def civick(self):
        try:
            return RaceEthReidVT(self.hci_map.get(int(self)))
        except (ValueError, KeyError) as e:
            return FailedValue(str(self), e)


class RaceEthCen00VT(RaceEthVT):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/cen00'

class RaceEthCen10VT(RaceEthVT):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/cen10'

class RaceEthOmbVT(RaceEthVT):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/omb'

class RaceEthReidVT(IntValue):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/civick'
    lom = LOM.NOMINAL

    @property
    def name(self):
        try:
            return RaceEthCivickNameVT(RaceEthVT.civick_map[int(self)][1])
        except Exception as e:
            return FailedValue(self, e)

class RaceEthCivickNameVT(TextValue):
    role = ROLE.DIMENSION
    vt_code = 'd/raceth/civick/name'
    lom = LOM.NOMINAL


class AgeVT(IntDimension):
    """A single-year age"""
    role = ROLE.DIMENSION
    vt_code = 'd/age'
    lom = LOM.ORDINAL

class AgeRangeVT(StrDimension):
    """An age range, between two years. The range is half-open. """
    role = ROLE.DIMENSION
    vt_code = 'd/age/range'
    lom = LOM.ORDINAL

    # Standard age ranges
    ranges = [
        (0,6),
        (6,18),
        (0,18),
        (0,21),
        (18,35),
        (18,65),
        (35,65),
        (65,85),
        (65,100),
        (85,100)
    ]

    def __init__(self, v):
        parts = v.split('-')
        self.from_year, self.to_year = int(parts[0]), int(parts[1])

class AgeRangeCensus(StrDimension):
    """Age ranges that appear in census column titles"""
    role = ROLE.DIMENSION
    vt_code = 'd/age/range/census'
    lom = LOM.ORDINAL

    under = re.compile('[Uu]nder (?P<to>\d+)')
    over = re.compile('(?P<from>\d+) years and over')
    to = re.compile('(?P<from>\d+) to (?P<to>\d+) years')
    _and = re.compile('(?P<from>\d+) and (?P<and2>\d+) years')
    one = re.compile('(?P<one>\d+) years')

    from_year = None
    to_year = None

    def parse_age_group(self,v):
        for rx in (self.under, self.over, self.to, self._and, self.one):
            m = rx.search(v)
            if m:
                d = m.groupdict()

                if rx == self.under:
                    d['from'] = 0
                elif rx == self.over:
                    d['to'] = 120
                elif rx == self._and:
                    d['to'] = int(d['and2']) + 1
                    del d['and2']
                elif rx == self.to:
                    d['to'] = int(d['to']) + 1
                elif rx == self.one:
                    d['from'] = d['one']
                    d['to'] = d['one']
                    del d['one']

                return int(d['from']), int(d['to'])

        return None, None

    def __init__(self, v):
        self.from_year, self.to_year = self.parse_age_group(v)

    def __str__(self):
        return "{:02d}-{:02d}".format(self.from_year, self.to_year)

class Decile(IntDimension):
    """A Decile Ranking, from 1 to 10"""
    role = ROLE.DIMENSION
    vt_code = 'd'
    desc = "Decile ranking"
    lom = LOM.ORDINAL

class Quartile(IntDimension):
    """A Quartile Ranking, from 1 to 4"""
    role = ROLE.DIMENSION
    vt_code = 'd'
    desc = "Quartile ranking"
    lom = LOM.ORDINAL

class Quintile(IntDimension):
    """A Decile Ranking, from 1 to 10"""
    vt_code = 'd'
    desc = "Quintile ranking"
    lom = LOM.ORDINAL

class PercentileVT(FloatDimension):
    """Percentile ranking, 0 to 100 """
    vt_code = 'pctl'
    desc = 'Percentile Rank'

    def __new__(cls, v):

        if isinstance(v,text_type) and '%' in v:
            v = v.strip('%')

        return FloatDimension.__new__(cls, v)

    @property
    def rate(self):
        return float(self) / 100.0


dimension_value_types = {
    "key": KeyVT,
    "id": IdentifierVT,
    'dimension': StrDimension,
    'dimension/str': StrDimension,
    'dimension/text': TextDimension,
    'dimension/int': IntDimension,
    'label': LabelValue,
    'name/first': StrDimension,
    'name/last': StrDimension,
    'name/middle': StrDimension,
    "raceth": RaceEthVT,
    "raceth/hci": RaceEthCodeHCI,
    "raceth/cen00": RaceEthCen00VT,
    "raceth/cen10": RaceEthCen10VT,
    "raceth/omb": RaceEthOmbVT,
    "raceth/civick": RaceEthReidVT,
    "age": AgeVT,
    "age/range": AgeRangeVT, #age_range
    "decile": Decile,
    'quartile': Quartile,
    'quintile': Quintile,
    "pctl": PercentileVT,
    "percentile": PercentileVT
}



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""
Archive Urls represent an archive file in the local file system, such as a Zip file.
They provide implementations of :py:meth:`Url.get_target` that will extract the
target file from the archive and store it in the file cache, returning a new Url to
the extracted file."""


from .zip import ZipUrl

__all__ = ["ZipUrl"]


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from rowgenerators.generator.csv import CsvSource
from rowgenerators.source import Source


class GooglePublicSource(CsvSource):
    url_template = 'https://docs.google.com/spreadsheets/d/{key}/export?format=csv&gid={gid}'

    @classmethod
    def download_url(cls, spec):
        return cls.url_template.format(key=spec.netloc)



class GoogleAuthenticatedSource(Source):
    """Generate rows from a Google spreadsheet source that requires authentication

    To read a GoogleSpreadsheet, you'll need to have an account entry for google_spreadsheets, and the
    spreadsheet must be shared with the client email defined in the credentials.

    Visit http://gspread.readthedocs.org/en/latest/oauth2.html to learn how to generate the credential file, then
    copy the entire contents of the file into the a 'google_spreadsheets' key in the accounts file.

    Them share the google spreadsheet document with the email addressed defined in the 'client_email' entry of
    the credentials.

    """

    def __iter__(self):
        """Iterate over all of the lines in the file"""

        self.start()

        for row in self._fstor.get_all_values():
            yield row

        self.finish()




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from .file import FileUrl

class ProgramUrl(FileUrl):
    """URL that references an executable file"""

    match_priority = FileUrl.match_priority - 1

    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)
        kwargs['proto'] = 'program'

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'program'

    def get_resource(self):
        return super().get_resource()

    def get_target(self):
        return super().get_target()









# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from .file import FileUrl
from rowgenerators.exceptions import AppUrlError
class ExcelFileUrl(FileUrl):
    """URL that references an Excl file, either .xls or .xlsx"""

    match_priority = FileUrl.match_priority-5

    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)

        # Move the target_file to the sequence, if there is only one.
        if self._parts.get('target_file') and not self._parts.get('target_segment'):
            self._parts['target_segment'] = self._parts['target_file']
            del self._parts['target_file']



    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'file' and url.resource_format in ('xlsx', 'xls')

    @property
    def resource_url(self):
        from rowgenerators.appurl.util import unparse_url_dict

        return unparse_url_dict(self.dict,
                                scheme_extension=False,
                                fragment=False)

    @property
    def target_file(self):
        return self.resource_file

    # Just a copy of the one from Url; looks like it must be reset because
    # the target_file prop was replaced
    @target_file.setter
    def target_file(self, v):
        raise NotImplementedError()

    @property
    def target_format(self):
        return 'xlsx'

    def list(self, list_self=False):

        from xlrd import open_workbook

        wb = open_workbook(filename=str(self.fspath))

        def _l():
            return list(self.set_target_segment(sheet) for sheet in wb.sheet_names())

        return ( [self] if list_self else [] ) + _l()



    def join(self, s):
        return super().join(s)

    def join_dir(self, s):
        return super().join_dir(s)

    def join_target(self, tf):

        u = self.clone()

        try:
            u.target_segment = tf.path
        except AttributeError:
            u.target_segment = tf

        return u

    def get_target(self):
        return self # Like super method, but don't clear the fragment; it's needed in the row generator











# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""
WebUrls represent resources that are accessible on the web. They provide implementations of
:py:meth:`Url.get_resource` that will download a file from the web and store it in the file
cache, returning a new Url object pointing to the downloaded file.
"""

from .download import Downloader
from .ckan import CkanUrl
from .google import GoogleProtoCsvUrl, GoogleSpreadsheetUrl
from .s3 import S3Url
from .socrata import SocrataUrl
from .web import WebUrl, FtpUrl
from .webpage import WebPageUrl



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """


class DataRowGenerator(object):
    """Returns only rows between the start and end lines, inclusive """

    def __init__(self, seq, start=0, end=None, **kwargs):
        """
        An iteratable wrapper that coalesces headers and skips comments

        :param seq: An iterable
        :param start: The start of data row
        :param end: The last row number for data
        :param kwargs: Ignored. Sucks up extra parameters.
        :return:
        """

        self.iter = iter(seq)
        self.start = start
        self.end = end
        self.headers = []  # Set externally

        int(self.start)  # Throw error if it is not an int
        assert self.start > 0

    def __iter__(self):

        for i, row in enumerate(self.iter):

            if i < self.start or (self.end is not None and i > self.end):
                continue

            yield row




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" Code generation for processing columns

"""

import ast


def file_loc():
    """Return file and line number"""
    import sys
    import inspect
    try:
        raise Exception
    except:
        file_ = '.../' + '/'.join((inspect.currentframe().f_code.co_filename.split('/'))[-3:])
        line_ = sys.exc_info()[2].tb_frame.f_back.f_lineno
        return "{}:{}".format(file_, line_)


const_args = ('row', 'row_n', 'scratch', 'errors', 'accumulator', 'pipe', 'manager', 'source')
var_args = ('v', 'i_s', 'i_d', 'header_s', 'header_d')
all_args = var_args + const_args

# Full lambda definition for a column, including variable parts
col_code_def = 'lambda {}:'.format(','.join(all_args))

# lambda definition for the who;e row. Includes only the arguments
# that are the same for every column
code_def = 'lambda {}:'.format(','.join(const_args))

col_args_t = """col_args = dict(v=v, i_s=i_s, i_d=i_d, header_s=header_s, header_d=header_d,
              scratch=scratch, errors=errors, accumulator = accumulator,
              row=row, row_n=row_n)"""

file_header = """
# {}
import sys
from rowgenerators.valuetype import resolve_value_type
from rowgenerators.rowpipe.exceptions import CasterExceptionError
""".format(file_loc())

column_template = """
# {loc}
def {f_name}(v, i_s, i_d, header_s, header_d, row, row_n, errors, scratch, accumulator, pipe, manager, source):

    try:
{stack}

    except Exception as exc:

{exception}

    return v
"""

indent = '        '

row_template = """
# {loc}
def row_{table}_{stage}(row, row_n, errors, scratch, accumulator, pipe, manager, source):

    return [
{stack}
    ]
"""




class CodeGenError(Exception):
    pass


def exec_context(**kwargs):
    """Base environment for evals, the stuff that is the same for all evals. Primarily used in the
    Caster pipe"""
    import dateutil.parser
    import datetime
    import random
    from functools import partial
    from rowgenerators.valuetype.types import parse_date, parse_time, parse_datetime
    import rowgenerators.valuetype.types
    import rowgenerators.valuetype.exceptions
    import rowgenerators.valuetype.test
    import rowgenerators.valuetype
    import rowgenerators.rowpipe.transforms

    def set_from(f, frm): # This maybe isn't used anymore, or maybe it is for debugging?
        try:
            try:
                f.ambry_from = frm
            except AttributeError:  # for instance methods
                f.im_func.ambry_from = frm
        except (TypeError, AttributeError):  # Builtins, non python code
            pass

        return f

    test_env = dict(
        parse_date=parse_date,
        parse_time=parse_time,
        parse_datetime=parse_datetime,
        partial=partial
    )

    test_env.update(kwargs)
    test_env.update(dateutil.parser.__dict__)
    test_env.update(datetime.__dict__)
    test_env.update(random.__dict__)
    test_env.update(rowgenerators.valuetype.core.__dict__)
    test_env.update(rowgenerators.valuetype.types.__dict__)
    test_env.update(rowgenerators.valuetype.exceptions.__dict__)
    test_env.update(rowgenerators.valuetype.test.__dict__)
    test_env.update(rowgenerators.valuetype.__dict__)
    test_env.update(rowgenerators.rowpipe.transforms.__dict__)

    localvars = {}

    for f_name, func in test_env.items():
        if not isinstance(func, (str, tuple)):
            localvars[f_name] = set_from(func, 'env')

    # The 'b' parameter of randint is assumed to be a manager, but
    # replacing it with a lambda prevents the param assignment
    localvars['randint'] = lambda a, b: random.randint(a, b)

    localvars['round'] = lambda a, b: round(a, b)

    return localvars


def make_row_processors(source_headers, dest_table, env=None):
    """
    Make multiple row processors for all of the columns in a table.

    :param source_headers:
    :param dest_table:
    :param env:

    :return:
    """

    import re

    if env is None:
        env = exec_context()

    assert len(dest_table.columns) > 0

    # Convert the transforms to a list of list, with each list being a
    # segment of column transformations, and each segment having one entry per column.

    row_processors = []

    out = []
    preamble = []

    transforms = dest_table.stage_transforms

    for i, segments in enumerate(transforms):  # Iterate over each stage

        column_names = []
        column_types = []
        seg_funcs = []

        # Iterate over each column, linking it to the segments for this stage
        for col_num, (segment, column) in enumerate(zip(segments, dest_table), 0):

            assert column
            assert column.name, (dest_table.name, i)
            assert column.name == segment['column'].name

 
            col_name = column.name
            preamble_parts, try_lines, exception, passthrough = make_stack(env, i, segment)

            preamble += preamble_parts

            column_names.append(col_name)
            column_types.append(column.datatype)

            # Optimization to remove unecessary functions. Without this, the column function will
            # have just 'v = v'
            if len(segment['transforms']) == 1 and segment['transforms'][0] == 'v':
                seg_funcs.append('row[{}]'.format(col_num))
                continue

            column_name = re.sub(r'[^\w]+', '_', col_name, )
            table_name = re.sub(r'[^\w]+', '_', dest_table.name)

            assert column_name, (dest_table.name, i, col_name)
            assert table_name

            f_name = "{}_{}_{}".format(table_name, column_name, i)

            exception = (exception if exception
                         else 'raise CasterExceptionError("' + f_name + '",header_d, v, exc, sys.exc_info())')

            try:
                # The input values for the first stage is the input dataset,
                # which may have different columns that the later stages
                if i == 0:
                    i_s = source_headers.index(column.name)
                    header_s = column.name

                else:
                    i_s = col_num
                    header_s = None

                v = 'row[{}]'.format(i_s)

            except ValueError as e:  # The col name is not in the source dataset
                # This is the typical case when the output dataset has different columns from the
                # input, whcih should only occur on the first stage.
                i_s = 'None'
                header_s = None
                v = 'None' if col_num >= 1 else 'row_n'  # Give the id column (first column) the row number

            header_d = column.name

            # Seg funcs holds the calls to the function for each column, called in the row stage function
            seg_funcs.append(f_name
                             + ('({v}, {i_s}, {i_d}, {header_s}, \'{header_d}\', '
                                'row, row_n, errors, scratch, accumulator, pipe, manager, source)')
                             .format(v=v,
                                     i_s=i_s,
                                     i_d=col_num,
                                     header_s="'" + header_s + "'" if header_s else 'None',
                                     header_d=header_d))

            # This creates the column manipulation function.
            out.append(column_template.format(
                f_name=f_name,
                table_name=dest_table.name,
                column_name=col_name,
                stage=i,
                i_s=i_s,
                i_d=col_num,
                header_s=header_s,
                header_d=header_d,
                v=v,
                exception=indent + exception,
                stack='\n'.join(indent + l for l in try_lines),
                col_args='',  # col_args not implemented yet
                loc=file_loc()))

        # This stack assembles all of the function calls that will generate the next row
        stack = '\n'.join("{}{}, # column {}".format(indent, l, cn)
                          for l, cn, dt in zip(seg_funcs, column_names, column_types))

        out.append(row_template.format(
            table=re.sub(r'[^\w]+', '_', dest_table.name),
            stage=i,
            stack=stack,
            loc=file_loc()
        ))

        row_processors.append('row_{table}_{stage}'.format(stage=i,
                                                           table=re.sub(r'[^\w]+', '_', dest_table.name)))

    # Add the final datatype cast, which is done seperately to avoid an unecessary function call.
    stack = '\n'.join("{}cast_{}(row[{}], '{}', errors),".format(indent, c.datatype.__name__, i, c.name)
                      for i, c in enumerate(dest_table))

    out.append(row_template.format(
        table=re.sub(r'[^\w]+', '_', dest_table.name),
        stage=len(transforms),
        stack=stack,
        loc=file_loc()
    ))

    row_processors.append('row_{table}_{stage}'.format(stage=len(transforms),
                                                       table=re.sub(r'[^\w]+', '_', dest_table.name)))

    out.append('row_processors = [{}]'.format(','.join(row_processors)))

    return '\n'.join([file_header] + list(set(preamble)) + out)


def calling_code(f, f_name=None, raise_for_missing=True):
    """Return the code string for calling a function. """
    import inspect
    from rowgenerators.exceptions import ConfigurationError

    if inspect.isclass(f):
        try:
            args = inspect.signature(f.__init__).parameters.keys()
        except TypeError as e:
            raise TypeError("Failed to inspect {}: {}".format(f, e))

    else:
        args = inspect.signature(f).parameters.keys()

    if len(args) > 1 and list(args)[0] == 'self':
        args = list(args)[1:]

    if 'self' in args:  # Python3 gets self, but not Python2
        args.remove('self')

    for a in args:
        if a not in all_args + ('exception',):  # exception arg is only for exception handlers
            if raise_for_missing:
                # In CPython, inspecting __init__ for IntMeasure, FloatMeasure, etc,
                # raises a TypeError 12 lines up, but that does not happen in PyPy. This hack
                # raises the TypeError.
                if a == 'obj':
                    raise TypeError()

                raise TypeError()
                raise ConfigurationError('Caster code {} has unknown argument '
                                         'name: \'{}\'. Must be one of: {} '.format(f, a, ','.join(all_args)))

    arg_map = {e: e for e in var_args}

    args = [arg_map.get(a, a) for a in args]

    return "{}({})".format(f_name if f_name else f.__name__, ','.join(args))

def make_stack(env, stage, segment):
    """For each transform segment, create the code in the try/except block with the
    assignements for pipes in the segment """

    import string
    import random
    from rowgenerators.valuetype import ValueType

    passthrough = False  # If true, signal that the stack will just return its input value

    column = segment['column']

    def make_line(column, t):
        preamble = []

        line_t = "v = {} # {}"

        env_t = env.get(t,t)

        if isinstance(env_t, type) and issubclass(env_t, ValueType):  # A valuetype class, from the datatype column.

            try:
                cc, fl = calling_code(env_t, env_t.__name__), file_loc()
            except TypeError as e:
                cc, fl = "{}(v)".format(env_t.__name__), file_loc()

            preamble.append("{} = resolve_value_type('{}') # {}".format(env_t.__name__, env_t.vt_code, file_loc()))

        elif isinstance(t, type):  # A python type, from the datatype columns.
            cc, fl = "parse_{}(v, header_d)".format(t.__name__), file_loc()

        elif callable(env.get(t)):  # Transform function
            try:
                cc, fl = calling_code(env.get(t), t), file_loc()
            except TypeError as e:
                raise

        else:  # A transform generator, or python code.

            rnd = (''.join(random.choice(string.ascii_lowercase) for _ in range(6)))

            name = 'tg_{}_{}_{}'.format(column.name, stage, rnd)
            try:
                a, b, fl = rewrite_tg(env, name, t)
            except (CodeGenError, AttributeError) as e:
                raise CodeGenError("Failed to re-write pipe code '{}' in column '{}': {} "
                                   .format(t, column, e))

            cc = str(a)

            if b:
                preamble.append("{} = {} # {}".format(name, b, file_loc()))

        line = line_t.format(cc, fl)

        return line, preamble

    preamble = []

    try_lines = []


    for t in list(segment):

        if not t:
            continue

        line, col_preamble = make_line(column, t)

        preamble += col_preamble
        try_lines.append(line)

    if segment['exception']:
        exception, col_preamble = make_line(column, segment['exception'])
    else:
        exception = None

    if len(try_lines) == 0:
        try_lines.append('pass # Empty pipe segment')

    assert len(try_lines) > 0, column.name

    return preamble, try_lines, exception, passthrough


def mk_kwd_args(fn, fn_name=None):
    import inspect

    fn_name = fn_name or fn.__name__

    fn_args = inspect.getargspec(fn).args

    if len(fn_args) > 1 and fn_args[0] == 'self':
        args = fn_args[1:]

    kwargs = dict((a, a) for a in all_args if a in args)

    return "{}({})".format(fn_name, ','.join(a + '=' + v for a, v in kwargs.items()))


class ReplaceTG(ast.NodeTransformer):
    """Replace a transform generator with the transform function"""

    def __init__(self, env, tg_name):
        super(ReplaceTG, self).__init__()

        self.tg_name = tg_name
        self.trans_gen = None
        self.env = env
        self.loc = ''

    def missing_args(self):
        pass

    def visit_Call(self, node):  # pragma: no cover

        import inspect
        from rowgenerators.valuetype import is_transform_generator
        import types

        if not isinstance(node.func, ast.Name):
            self.generic_visit(node)
            # Python 3.5 removed starargs, which the meta module, which dumps the AST to code,
            # still expects. Not sure about the need for kwargs, but meta expects that too
            node.starargs = None
            node.kwargs = None
            return node

        fn_name = node.func.id
        fn_args = None
        use_kw_args = True

        fn = self.env.get(node.func.id)
        self.loc = file_loc()  # Not a builtin, not a type, not a transform generator

        # In this case, the code line is a type that has a parse function, so rename it.
        if not fn:
            t_fn_name = 'parse_' + fn_name
            t_fn = self.env.get(t_fn_name)
            if t_fn:
                self.loc = file_loc()  # The function is a type
                fn, fn_name = t_fn, t_fn_name

        # Ok, maybe it is a builtin
        if not fn:
            o = eval(fn_name)
            if isinstance(o, types.BuiltinFunctionType):
                self.loc = file_loc()  # The function is a builtin
                fn = o
                fn_args = ['v']
                use_kw_args = False

        if not fn:
            raise CodeGenError("Failed to get function named '{}' from the environment".format(node.func.id))

        if not fn_args:
            fn_args = inspect.getargspec(fn).args

        # Create a dict of the arguments that have been specified
        used_args = dict(tuple(zip(fn_args, node.args))
                         + tuple((kw.arg, kw.value) for kw in node.keywords)
                         )

        # Add in the arguments that were not, but only for args that are specified to be
        # part of the local environment
        for arg in fn_args:
            if arg not in used_args and arg in all_args:
                used_args[arg] = ast.Name(id=arg, ctx=ast.Load())

        # Now, all of the args are in a dict, so we'll re-build them as
        # as if they were all kwargs. Any arguments that were not provided by the
        # signature in the input are added as keywords, with the value being
        # a variable of the same name as the argument: ie. if 'manager' was defined
        # but not provided, the signature has an added 'manager=manager' kwarg

        keywords = [ast.keyword(arg=k, value=v) for k, v in used_args.items()]

        tg_ast = ast.copy_location(
            ast.Call(
                func=ast.Name(id=fn_name, ctx=ast.Load()),
                args=[e.value for e in keywords] if not use_kw_args else [],  # For builtins, which only take one arg
                keywords=keywords if use_kw_args else [],
                starargs=[],
                kwargs=[]
            ), node)

        if is_transform_generator(fn):
            raise Exception("Deprecated?")
            self.loc = file_loc()  # The function is a transform generator.
            self.trans_gen = tg_ast
            replace_node = ast.copy_location(
                ast.Call(
                    func=ast.Name(id=self.tg_name, ctx=ast.Load()),
                    args=[],
                    keywords=[],
                    kwargs=ast.Name(id='col_args', ctx=ast.Load()),
                    starargs=[]
                ), node)

        else:

            replace_node = tg_ast


        return replace_node


def rewrite_tg(env, tg_name, code):
    """Re-write a transform generating function pipe specification by extracting the transform generating part,
    and replacing it with the generated transform. so:

       tgen(a,b,c).foo.bar

    becomes:

        tg = tgen(a,b,c)

        tg.foo.bar

    """

    import meta  # Byte-code and ast programming tools
    import codegen

    visitor = ReplaceTG(env, tg_name)
    assert visitor.tg_name

    try:
        tree = visitor.visit(ast.parse(code.strip()))
    except SyntaxError as e:
        raise SyntaxError(str(e) + "\nIn code: \n" + code)

    if visitor.loc:
        loc = ' #' + visitor.loc
    else:
        loc = file_loc()  # The AST visitor didn't match a call node

    if visitor.trans_gen:
        tg = meta.dump_python_source(visitor.trans_gen).strip()
    else:
        tg = None


    return meta.dump_python_source(tree).strip(), tg, loc



import unittest


def get_file(url_str):

    from rowgenerators import  parse_app_url

    u = parse_app_url(url_str)

    return u.get_resource().get_target()


def data_path(v):
    from os.path import dirname, join
    d = dirname(__file__)
    return join(d, 'test_data', v)


def script_path(v=None):
    from os.path import dirname, join
    d = dirname(__file__)
    if v is not None:
        return join(d, 'scripts', v)
    else:
        return join(d, 'scripts')


def sources():
    import csv
    with open(data_path('sources.csv')) as f:
        r = csv.DictReader(f)
        return list(r)


class RowGeneratorTest(unittest.TestCase):
    """Test Metapack AppUrls and Row Generators"""

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from os.path import basename, join
from rowgenerators.appurl.web.web import WebUrl

class SocrataUrl(WebUrl):
    """Url to represent a dataset stored in a Socrata data repository. """

    def __init__(self, url=None,downloader=None, **kwargs):

        kwargs['resource_format'] = 'csv'
        kwargs['encoding'] = 'utf8'
        kwargs['proto'] = 'socrata'

        super().__init__(url,downloader=downloader, **kwargs)



    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'socrata'


    @property
    def resource_url(self):
        from rowgenerators.appurl.util import unparse_url_dict

        return unparse_url_dict(self.__dict__,
                                scheme_extension=False,
                                fragment=False,
                                path=join(self.path, 'rows.csv'))

    @property
    def resource_file(self):
        return basename(self.path)+'.csv'



    def get_resource(self):
        """Get the contents of resource and save it to the cache, returning a file-like object"""
        raise NotImplementedError()

    def get_target(self):
        """Get the contents of the target, and save it to the cache, returning a file-like object
        :param downloader:
        """
        raise NotImplementedError()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from os.path import normpath, join, exists
from os import environ
import json

from rowgenerators.exceptions import SourceError
from rowgenerators.source import Source

class ProgramSource(Source):
    """Generate rows from a program. Takes kwargs from the spec to pass into the program. """

    def __init__(self, ref, cache=None, working_dir=None, env = None, **kwargs):

        super().__init__(ref, cache, working_dir, **kwargs)

        import platform

        if platform.system() == 'Windows':
            raise NotImplementedError("Program sources aren't working on Windows")

        assert working_dir

        self.program = normpath(join(working_dir, self.ref.path))

        if not exists(self.program):
            raise SourceError("Program '{}' does not exist".format(self.program))

        self.options = []

        self.properties = {}

        self.env = {}

        # Expand the generator args and kwargs into parameters for the program,
        # which may be command line options, env vars, or a json encoded dict in the PROPERTIES
        # envvar.
        for k, v in (env or {}).items():
            if k.startswith('--'):
                # Long options
                self.options.append("{} {}".format(k,v))
            elif k.startswith('-'):
                # Short options
                self.options.append("{} {}".format(k, v))
            elif k == k.upper():
                #ENV vars
                self.env[k] = v
            else:
                # Normal properties, passed in as JSON and as an ENV
                self.properties[k] = v

        self.env['PROPERTIES'] = json.dumps(self.properties)

        # Make sure that sys.stdout is always UTF*. It can end up US_ASCI otherwise.
        self.env['PYTHONIOENCODING']='utf-8:replace'

    def start(self):
        pass

    def finish(self):
        pass

    def open(self):
        pass

    def __iter__(self):
        import csv
        import subprocess
        from io import TextIOWrapper

        import sys


        if self.program.endswith('.py'):
            # If it is a python program, it's really nice, possibly required,
            # that the program be run with the same interpreter as is running this program.
            #
            # The -u option makes output unbuffered.  http://stackoverflow.com/a/17701672
            prog = [sys.executable, '-u', self.program]
        else:
            prog = [self.program]


        p = subprocess.Popen(prog + self.options,
                             stdout=subprocess.PIPE,
                             bufsize=1,
                             env=self.env)

        yield from csv.reader(TextIOWrapper(p.stdout, encoding='utf8', errors='replace'))



from __future__ import print_function

import unittest

from rowgenerators.rowpipe.pipeline import Pipeline, Pipe, PrintRows, Sample, Head, SelectRows, Slice


def cast_str(v):
    return str(v)


def cast_int(v):
    return int(v)


def cast_float(v):
    return float(v)


class Test(unittest.TestCase):

    def test_sample_head(self):

        class Source(Pipe):

            def __iter__(self):
                yield ['int', 'int']

                for i in range(10000):
                    yield ([i, i])

        # Sample
        pl = Pipeline(
            source=Source(),
            first=Sample(est_length=10000),
            last=PrintRows(count=50)
        )

        pl.run()

        # head
        self.assertIn([7, 7], pl[PrintRows].rows)
        self.assertIn([2018, 2018], pl[PrintRows].rows)
        self.assertIn([9999, 9999], pl[PrintRows].rows)

        pl = Pipeline(
            source=Source(),
            first=Head(10),
            last=PrintRows(count=50)
        )

        pl.run()

        self.assertEquals(10, len(pl[PrintRows].rows))

        print(pl)

    def test_select(self):

        class Source(Pipe):
            def __iter__(self):
                yield ['a', 'b']

                for i in range(10000):
                    yield ([i, i])

        pl = Pipeline(
            source=Source(),
            first=SelectRows('row.a == 100 or row.b == 1000'),
            last=PrintRows(count=50)
        )

        pl.run()

        rows = pl[PrintRows].rows

        self.assertEqual(2, len(rows))
        self.assertEqual(100, rows[0][0])
        self.assertEqual(1000, rows[1][1])

    def test_slice(self):

        self.assertEquals('lambda row: tuple(row[0:3])+tuple(row[10:13])+(row[9],)+(row[-1],)',
                          Slice.make_slicer((0, 3), (10, 13), 9, -1)[1])

        self.assertEquals('lambda row: tuple(row[0:3])+tuple(row[10:13])+(row[9],)+(row[-1],)',
                          Slice.make_slicer("0:3,10:13,9,-1")[1])

        return

        class Source(Pipe):
            def __iter__(self):
                yield ['col' + str(j) for j in range(20)]

                for i in range(10000):
                    yield [j for j in range(20)]

        # Sample
        pl = Pipeline(
            source=[Source(), Slice((0, 3), (10, 13), 9, -1)],
            last=PrintRows(count=50)
        )

        pl.run()

        self.assertEquals(
            [1, 0, 1, 2, 10, 11, 12, 9, 19],
            pl[PrintRows].rows[0])
        self.assertEquals(
            ['col0', 'col1', 'col2', 'col10', 'col11', 'col12', 'col9', 'col19'],
            pl[PrintRows].headers)

        self.assertEqual(
            [('0', '3'), ('10', '13'), 9, -1],
            Slice.parse("0:3,10:13,9,-1"))

        pl = Pipeline(
            source=[Source(), Slice("0:3,10:13,9,-1")],
            last=PrintRows(count=50)
        )

        pl.run()

        self.assertEquals([1, 0, 1, 2, 10, 11, 12, 9, 19], pl[PrintRows].rows[0])
        self.assertEquals(
            ['col0', 'col1', 'col2', 'col10', 'col11', 'col12', 'col9', 'col19'],
            pl[PrintRows].headers)

    def test_multi_source(self):

        class Source(Pipe):

            def __init__(self, start):
                self.start = start

            def __iter__(self):

                for i in range(self.start, self.start + 10):
                    if i == self.start:
                        yield ['int', 'int']  # header

                    yield ([self.start, i])

            def __str__(self):
                return 'Source {}'.format(self.start)

        # Sample
        pl = Pipeline(last=PrintRows(count=50))

        pl.run(source_pipes=[Source(0), Source(10), Source(20)])

        self.assertIn([0, 2], pl[PrintRows].rows)
        self.assertIn([10, 18], pl[PrintRows].rows)
        self.assertIn([20, 21], pl[PrintRows].rows)



# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""

Create row processor for generating JSON

The format for each line is a sequence of path elements, with a terminal at the end. For instance:

    a.b.c.t


Assume there is a path specification for each column in the table, and that the column name specification substituted
for a missing path specification.

A path element without a '[]' specifies a key to an object, and a path element with a '[]' specifies an array.
A path element with a '[.]' specifies that the value should be set on the same list item as the last path element.

* "a: v": Create Key a, set to v
* "a.b: v": Create Key a, set to new object. Create key b, set to v
* "a.b[]: v" Create key a, set to new object. Create key b, set to new list. Append v
* "a[].b: v" Create key a, set to new list. Create new object. Create key b, set to v

The "[-]" means to use the last element of the existing list.

Types of path elements:

    a: terminal, object key
    a.: nonterminal, object key and new object
    a[]: terminal, add to new or existing list
    a[].: nonterminal, new array with new object
    a[-].:nonterminal, last object in list

"""

import json

def parse_path(path):
    """Decompose a JSON structure path"""

    parts = []

    for e in path.split('.'):
        if e.endswith('[]'):
            k,_ = e.split('[')
            parts.append([k,'an', False])  # array, new
        elif e.endswith('[-]'):
            k, _ = e.split('[')
            parts.append([k, 'al', False]) # array, last
        else:
            k = e
            parts.append([k, 'o', False])  # object

    if parts:
        parts[-1][2] = True # The last item is the terminal

    return parts

def add_to_struct(s, path, v):
    """Add a value v into a complex data structure s at a given path. """
    o = s

    path_parts = parse_path(path)

    #print("----", path)

    for i,  (key, type, is_terminal) in enumerate(path_parts):

        if type == 'an' and not is_terminal:

            if key not in o:
                o[key] = []

            o[key].append({})

            o = o[key][-1]

        elif type == 'al' and not is_terminal:
            if key not in o:
                raise Exception("Expected list '{}' to exist in '{}' ".format(key, path) )

            o = o[key][-1]

        elif type == 'o' and not is_terminal:
            if key not in o:
                o[key] = {}
            o = o[key]
        elif type == 'an' and is_terminal:
            if key not in o:
                o[key] = []
            o[key].append(v)

        elif type == 'al' and is_terminal:
            print("E")

        elif type == 'o' and is_terminal:
            o[key] = v
        else:
            print("G ", type, is_terminal)

class VTEncoder(json.JSONEncoder):
     """A JSON object encoder that can handle dates and times. Just converts them to strings"""

     def default(self, obj):
         from rowgenerators.valuetype import DateTimeVT, DateVT, TimeVT

         if isinstance(obj, (DateTimeVT, DateVT, TimeVT) ):
             return str(obj)

         # Let the base class default method raise the TypeError
         try:
            return json.JSONEncoder.default(self, obj)
         except TypeError as e:
             raise TypeError(DateTimeVT,type(obj))


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from rowgenerators.appurl import Url


class CkanUrl(Url):
    def __init__(self, url, **kwargs):
        raise NotImplementedError()
        kwargs['proto'] = 'ckan'
        super(CkanUrl, self).__init__(url, **kwargs)






# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
from rowgenerators.exceptions import SourceError, RowGeneratorError
from rowgenerators.source import Source
from rowgenerators.appurl import parse_app_url
import pandas as pd
import numpy as np


def iterate_pandas(df):


    if len(df.index.names) == 1 and df.index.names[0] is None and df.index.dtype != np.dtype('O'):
        # For an unnamed, single index, assume that it is just a row number
        # and we don't really need it

        yield list(df.columns)

        for index, row in df.iterrows():
            yield list(row)

    else:

        # Otherwise, either there are more than

        index_names = [n if n else "index{}".format(i) for i, n in enumerate(df.index.names)]

        yield index_names + list(df.columns)

        if len(df.index.names) == 1:
            idx_list = lambda x: [x]
        else:
            idx_list = lambda x: list(x)

        for index, row in df.iterrows():
            yield idx_list(index) + list(row)

def to_codes(df):
    """Return a dataframe with all of the categoricals represented as codes"""
    df = df.copy()
    cat_cols = df.select_dtypes(['category']).columns
    df[cat_cols] = df[cat_cols].apply(lambda x: x.cat.codes)
    return df

def make_cat_map(df, column):
    """Extract the mapping between category names and codes from the dataset. There should be an interface
    on the categorical index type for this map, but I could not find it. """

    t = pd.DataFrame( {'codes': df[column].cat.codes, 'values': df[column]} ).drop_duplicates().sort_values('codes')
    return { e.codes:e.values for e in list(t.itertuples())}

def extract_categories(fspath):
    """
    Create a Metatab doc with the schema for a CHIS file, including the categorical values.

    :param fspath: Path to the stata file.
    :return:
    """

    dfs = pd.read_stata(fspath)
    itr = pd.read_stata(fspath, iterator=True)

    var_d = itr.variable_labels()

    # We ought to be able to just use the value_labels() method to get value labels, but
    # there are a lot of variables it doesn't return values for. For those, we'll use
    # make_cat_map, which analyzes the categorical values directly.

    columns = []
    for col_name in dfs.columns:

        col = {
            'name' : col_name,
            'description':var_d[col_name],
            'ordered': False,
            'values': {}
        }

        try:
            col['values'] = make_cat_map(dfs, col_name)
        except AttributeError as e:
            # Probably, column is not a category
            pass

        try:
            if dfs[col_name].cat.ordered:
                col['ordered'] = True
        except AttributeError as e:
            pass

        columns.append(col)

    return columns

class StataSource(Source):
    """Generate rows from a stata object """

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        """

        Setting a fragment query with a value of 'values=codes' will
        cause data frame to iterate over codes, rather than categoricals.

        :param ref:
        :param cache:
        :param working_dir:
        :param env:
        :param kwargs:
        """

        super().__init__(ref, cache, working_dir, **kwargs)

        if not working_dir in sys.path:
            sys.path.append(working_dir)

        self.env = env or {}

        source_url = kwargs.get('source_url')

        u_ = parse_app_url(source_url)
        self.target_file = u_.target_file
        self.target_segment = u_.target_segment

        self.value_type = u_.fragment_query.get('values', 'categorical')

        self.kwargs = kwargs

    def dataframe(self, *args, **kwargs):
        """Return a pandas dataframe from the resource"""

        import pandas as pd

        for k,v in tuple(kwargs.items()):
            if k not in ('convert_dates', 'convert_categoricals', 'index_cols',
                         'convert_missing', 'preserve_dtypes','columns',
                         'order_categoricals', 'chunksize'):
                del kwargs[k]

        return pd.read_stata(self.ref.fspath, *args, **kwargs)

    @property
    def columns(self):
        import pandas as pd

        return extract_categories(self.ref.fspath)


    def __iter__(self):

        df = self.dataframe()

        if self.value_type == 'codes':
            df = to_codes(df)

        yield from iterate_pandas(df)



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
import types
from itertools import islice

from rowgenerators.exceptions import RowGeneratorError
from rowgenerators.source import Source
from rowgenerators.appurl.util import import_name_or_class


class PythonSource(Source):
    """Generate rows from a callable object. Takes kwargs from the spec to pass into the program.
    If the callable returns a pandas dataframe, iterate it as a PandasDataframeSource """

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):

        super().__init__(ref, cache, working_dir, **kwargs)

        if not working_dir in sys.path:
            sys.path.append(working_dir)

        self.env = env or {}

        self.kwargs = kwargs

    @property
    def object(self):
        """Return the python thing, a class or a function, that will be invoked """

        if not self.ref.path and self.ref.target_file in self.env:
            return self.env[self.ref.target_file]

        try:
            path = self.ref.path.replace('/', '.') + '.' + self.ref.target_file
            return import_name_or_class(path)

        except ModuleNotFoundError:
            pass

        if self.ref.fspath.parent not in sys.path:
            sys.path.insert(0, str(self.ref.fspath.parent))

        mod = __import__(self.ref.fspath.stem)

        return getattr(mod, self.ref.target_file)

    def __call__(self):

        self.kwargs.update(self.ref.fragment_query)

        '''Call the referenced function and return the result '''
        return self.object(env=self.env, cache=self.cache, **self.kwargs)

    def dataframe(self, *args, **kwargs):

        from pandas import DataFrame

        o = self()

        if isinstance(o, types.GeneratorType):

            # Just normal data, so use the iterator in this object.
            headers = next(islice(self, 0, 1))
            data = islice(self, 1, None)

            return DataFrame(list(data), columns=headers)

        elif isinstance(o, DataFrame):
            return o

        else:
            return super().dataframe(*args, **kwargs)

    def __iter__(self):
        """Iterate rows from the python generator returned from the referenced function. Or, if the
        function returns a dataframe, yield from a PandasDataframeSource constructed on the dataframe"""

        o = self()

        if isinstance(o, types.GeneratorType):

            try:
                yield from o
            except TypeError as e:
                # call to Python Url has wrong signature

                raise RowGeneratorError(str(e))

        else:
            from metapack import get_cache

            yield from PandasDataframeSource('<df>', o, get_cache())


class PandasDataframeSource(Source):
    """Iterates a pandas dataframe  """

    def __init__(self, url, df, cache, working_dir=None, **kwargs):
        super().__init__(url, cache, working_dir, **kwargs)

        self._df = df

    def __iter__(self):

        import numpy as np
        from itertools import islice

        self.start()

        df = self._df

        if len(df.index.names) == 1 and df.index.names[0] is None and df.index.dtype != np.dtype('O'):
            # For an unnamed, single index, assume that it is just a row number
            # and we don't really need it

            yield list(df.columns)

            for index, row in df.iterrows():
                yield list(row)

        else:

            # Otherwise, either there are more than

            index_names = [n if n else "index{}".format(i) for i, n in enumerate(df.index.names)]

            yield index_names + list(df.columns)

            if len(df.index.names) == 1:
                idx_list = lambda x: [x]
            else:
                idx_list = lambda x: list(x)

            for index, row in df.iterrows():
                yield idx_list(index) + list(row)

        self.finish()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from rowgenerators.appurl import Url

class WebPageUrl(Url):
    """A URL for webpages, not for data"""

    def __init__(self, url=None, downloader=None, **kwargs):
        raise NotImplementedError()
        super().__init__(url, downloader, **kwargs)



import unittest
from rowgenerators import get_generator
from rowgenerators import parse_app_url, get_cache

class TestIssues(unittest.TestCase):


    def x_test_pass_target_format(self):

        us = 'file:///Users/eric/Downloads/7908485365090507159.zip#VictimRecords.txt&target_format=csv'

        u = parse_app_url(us, target_format=None)

        print(u)
        r = u.get_resource()
        print(r)
        t = r.get_target()
        print(t)

        g = t.generator

        print(len(list(g)))

    def test_broken_zip(self):

        us = 'http://public.source.civicknowledge.com/example.com/sources/test_data.foo#simple-example.csv&resource_format=zip&encoding=ascii'

        u = parse_app_url(us, target_format=None)

        print(u)
        r = u.get_resource()
        print(r)
        t = r.get_target()
        print(t)

        g = t.generator

        print(len(list(g)))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """


from .iterator import IteratorSource

class GeneratorSource(IteratorSource):
    pass


"""


Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

from .core import *
import re
from datetime import date, time, datetime
from  dateutil.parser import parse

def cast_date(v, header_d, errors):

    if v is None or v is NoneValue or v == '':
        return None
    elif isinstance(v, FailedValue):
        pass
    elif isinstance(v, date):
        return v
    elif isinstance(v, ValueType):
        return v.__date__()
    elif isinstance(v, str):
        return parse(v).date

    errors[header_d].add(u"Failed to cast '{}' ({}) to date in '{}': {}".format(v, type(v), header_d, v.exc))
    count_errors(errors)
    return None


def cast_datetime(v, header_d, errors):

    if v is None or v is NoneValue or v == '':
        return None
    elif isinstance(v, datetime):
        return v
    elif isinstance(v, ValueType):
        return v.__datetime__()
    elif isinstance(v, str):
        return parse(v)
    elif isinstance(v, Exception):
        errors[header_d].add(u"Failed to cast '{}' ({}) to datetime in '{}': {}".format(v, type(v), header_d, v.exc))
        count_errors(errors)
    else:
        raise Exception("Unknown value! type={} value={}".format(type(v),v))

    return None


def cast_time(v, header_d, errors):
    if v is None or v is NoneValue or v == '':
        return None
    elif isinstance(v, time):
        return v
    elif isinstance(v, ValueType):
        return v.__time__()
    elif isinstance(v, str):
        return parse(v).time

    errors[header_d].add(u"Failed to cast '{}' ({}) to time in '{}': {}".format(v, type(v), header_d, v.exc))
    count_errors(errors)
    return None



class TimeVT(TimeValue, TimeMixin):
    role = ROLE.DIMENSION
    vt_code = 'dt/time'
    desc = 'Time'
    lom = LOM.ORDINAL


class DateVT(DateValue, TimeMixin):
    role = ROLE.DIMENSION
    vt_code = 'dt/date'
    desc = 'Date'
    lom = LOM.ORDINAL

    def __init__(self, v):
        super(DateVT, self).__init__(v)





class DateTimeVT(DateTimeValue, TimeMixin):
    role = ROLE.DIMENSION
    vt_code = 'dt/datetime'
    desc = 'Date and time'
    lom = LOM.ORDINAL



class YearValue(IntValue, TimeMixin):
    """Time interval of a single year"""
    role = ROLE.DIMENSION
    vt_code = 'year'
    desc = 'Single year Interval'
    lom = LOM.ORDINAL

    year_re = re.compile(r'^(\d{4})$')

    def __new__(cls, v):
        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        return IntValue.__new__(cls, v)

    @property
    def start(self):
        return int(self)

    @property
    def end(self):
        return int(self) + 1


class MonthValue(IntValue, TimeMixin):
    """A month"""
    role = ROLE.DIMENSION
    vt_code = 'month'
    desc = 'A month'
    lom = LOM.ORDINAL

class YearRangeValue(StrDimension, TimeMixin):
    """A half-open time interval between two years"""

    vt_code = 'year/range'
    desc = 'Year range interval'
    lom = LOM.ORDINAL

    y1 = None
    y2 = None

    year_range_re = re.compile(r'(\d{4})(?:\/|-|--)(\d{4})')  # / and -- make it also an ISO interval

    def __new__(cls, v):

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        o = StrValue.__new__(cls, v)

        if isinstance(v, (int, float)):
            o.y1 = v
            o.y2 = v + 1
            return o

        if isinstance(v, str):
            v = v.strip()

        m = YearValue.year_re.match(v)

        if m:
            o.y1 = int(m.group(1))
            o.y2 = o.y1 + 1

            return o

        m = YearRangeValue.year_range_re.match(v)

        if m:
            o.y1 = int(m.group(1))
            o.y2 = int(m.group(2))

            return o

        return FailedValue(v, ValueError("YearRangeValue failed to match year range"))

    @property
    def start(self):
        return int(self.y1)

    @property
    def end(self):
        return int(self.y2)

    def __str__(self):
        return str(self.y1) + '/' + str(self.y2)


class IntervalValue(StrDimension, TimeMixin):
    """A generic time interval. A single year, year range, or an ISO Interval"""
    vt_code = 'interval'
    desc = 'Time interval'
    lom = LOM.ORDINAL

    year_range_re = re.compile(r'(\d{4})(?:\/|-|--)(\d{4})')  # / and -- make it also an ISO interval

    def __new__(cls, v):

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        # P1Y1D: This is probably

        if isinstance(v, (int, float)):
            return IntervalIsoVT('{}/{}'.format(int(v), int(v) + 1))

        if isinstance(v, str):
            v = v.strip()

        m = YearValue.year_re.match(v)

        if m:
            return IntervalIsoVT('{}/{}'.format(int(m.group(1)), int(m.group(1)) + 1))

        m = cls.year_range_re.match(v)

        if m:
            return IntervalIsoVT('{}/{}'.format(m.group(1), m.group(2)))

        return IntervalIsoVT(v)


class IntervalIsoVT(StrValue, TimeMixin):
    role = ROLE.DIMENSION
    vt_code = 'duration/iso'
    desc = 'ISO FOrmat Interval'
    lom = LOM.ORDINAL

    interval = None

    def __new__(cls, *args, **kwargs):
        v = args[0]

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return None

        return StrValue.__new__(cls, *args, **kwargs)

    def __init__(self, v):
        import aniso8601

        self.interval = aniso8601.parse_interval(v)

    def __str__(self):
        return str(self.interval[0]) + '/' + str(self.interval[1])

    @property
    def start(self):
        return self.interval[0]

    @property
    def end(self):
        return self.interval[1]


times_value_types = {
    'date': DateVT,
    'datetime': DateTimeVT,
    'time': TimeVT,
    'year': YearValue,
    'month': MonthValue,
    'year/range': YearRangeValue,
    "interval": IntervalValue,
    "interval/iso": IntervalIsoVT,
}



from __future__ import print_function

import unittest
from csv import DictReader
import os
from os.path import exists, dirname

from rowgenerators.appurl.archive.zip import ZipUrl
from rowgenerators.appurl.url import Url, parse_app_url
from rowgenerators.appurl.web.download import Downloader
from rowgenerators.appurl.web.s3 import S3Url
from rowgenerators.appurl.web.web import WebUrl

from rowgenerators.appurl.test.support import data_path

class BasicTests(unittest.TestCase):

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')

    def compare_dict(self, name, a, b):
        from rowgenerators.util import flatten
        fa = set('{}={}'.format(k, v) for k, v in flatten(a));
        fb = set('{}={}'.format(k, v) for k, v in flatten(b));

        # The declare lines move around a lot, and rarely indicate an error
        fa = {e for e in fa if not e.startswith('declare=')}
        fb = {e for e in fb if not e.startswith('declare=')}

        errors = len(fa - fb) + len(fb - fa)

        if errors:
            print("=== ERRORS for {} ===".format(name))

        if len(fa - fb):
            print("In a but not b")
            for e in sorted(fa - fb):
                print('    ', e)

        if len(fb - fa):
            print("In b but not a")
            for e in sorted(fb - fa):
                print('    ', e)

        self.assertEqual(0, errors)


    def test_traits(self):

        u = Url("http://example.com/foo/bar.csv")

        self.assertEqual('http',u.proto)
        self.assertEqual('/foo/bar.csv', u.path)

        self.assertEqual('http://example.com/foo/bar.csv', str(u))

        u.path = '/bar/baz.csv'

        self.assertEqual('http://example.com/bar/baz.csv', str(u))



    def test_entry_points(self):

        self.assertIsInstance(parse_app_url('s3://bucket.com/foo/bar/baz.zip'), S3Url)
        self.assertIsInstance(parse_app_url('http://bucket.com/foo/bar/baz.zip'), WebUrl)
        self.assertIsInstance(parse_app_url('file://bucket.com/foo/bar/baz.zip'), ZipUrl)

    def test_entry_point_priorities(self):
        from pkg_resources import iter_entry_points

        eps = []

        for ep in iter_entry_points(group='appurl.urls'):

            try:
                n = ep.load().__name__
            except Exception as e:
                print("Failed for ", ep)
                raise

            eps.append(n)

        self.assertIn('ExcelFileUrl', eps)
        self.assertIn('WebUrl', eps)


    def test_download(self):
        """Test all three stages of a collection of downloadable URLS"""

        dldr = Downloader.get_instance()

        with open(data_path('sources.csv')) as f:
            for e in DictReader(f):

                if e['name'] != 'mz_no_zip':
                    continue

                #print(e['name'])
                if not e['resource_class']:
                    continue


                u = parse_app_url(e['url'], downloader=dldr)

                self.assertEqual(e['url_class'], u.__class__.__name__, e['name'])
                self.assertEqual(e['resource_format'], u.resource_format, e['name'])

                r = u.get_resource()
                self.assertEqual(e['resource_class'],r.__class__.__name__, e['name'])
                self.assertEqual(e['resource_format'], r.resource_format, e['name'])

                t = r.get_target()

                self.assertEqual(e['target_class'], t.__class__.__name__, e['name'])
                self.assertEqual(e['target_format'], t.target_format, e['name'])
                self.assertTrue(exists(t.fspath))

    def test_url_classes(self):

        from rowgenerators.appurl import match_url_classes

        with open(data_path('url_classes.csv')) as f:
            for i, e in enumerate(DictReader(f),2):

                row = 'Row #{}: Url {}'.format(i, e['in_url'])

                if not e['class']:
                    continue

                u = parse_app_url(e['in_url'])

                self.assertEqual(e['url'], str(u), row )
                self.assertEqual(e['resource_url'], str(u.resource_url), row)
                self.assertEqual(e['resource_file'], u.resource_file, row)
                self.assertEqual(e['target_file'], u.target_file or '', row)

    def test_component(self):

        with open(data_path('components.csv')) as f:
            for i, e in enumerate(DictReader(f),2):

                b = parse_app_url(e['base_url'])
                c = parse_app_url(e['component_url'])

                self.assertEqual(e['class'], b.__class__.__name__, e['base_url'])
                self.assertEqual(e['join_dir'], str(b.join_dir(c)), e['base_url'])
                self.assertEqual(e['join'], str(b.join(c)), e['base_url'])
                self.assertEqual(str(e['join_target']), str(b.join_target(c)), "{} {} {}".format(i,e['base_url'],
                                                                                                 type(b)))


    def test_base_url(self):
        """Simple test of splitting and recombining"""

        for u_s in ('http://server.com/a/b/c/file.csv','http://server.com/a/b/c/file.csv#a',
                    'http://server.com/a/b/c/file.csv#a;b', 'http://server.com/a/b/c/archive.zip#file.csv'):
            self.assertEqual(u_s, str(Url(u_s)))

        self.assertEqual('file.csv', Url('http://server.com/a/b/c/file.csv').target_file)
        self.assertEqual('file.csv', Url('http://server.com/a/b/c/file.csv').resource_file)
        self.assertEqual('http://server.com/a/b/c/file.csv', Url('http://server.com/a/b/c/file.csv').resource_url)

        self.assertEqual('file.csv', Url('http://server.com/a/b/c/resource.zip#file.csv').target_file)
        self.assertEqual('resource.zip', Url('http://server.com/a/b/c/resource.zip#file.csv').resource_file)


    def test_parse_file_urls(self):
        urls = [
            ('file:foo/bar/baz', 'foo/bar/baz', 'file:foo/bar/baz'),
            ('file:/foo/bar/baz', '/foo/bar/baz', 'file:///foo/bar/baz'),
            ('file://example.com/foo/bar/baz', '/foo/bar/baz', 'file://example.com/foo/bar/baz'),
            ('file:///foo/bar/baz', '/foo/bar/baz', 'file:///foo/bar/baz'),
        ]

        for i, o, u in urls:
            p = parse_app_url(i)
            self.assertEqual(o, p.path)

            self.assertEqual(u, str(p))

        urls = [
            ('foo/bar/baz', 'foo/bar/baz', 'file:foo/bar/baz'),
            ('/foo/bar/baz', '/foo/bar/baz', 'file:///foo/bar/baz'),
        ]

        for i, o, u in urls:
            p = parse_app_url(i)
            self.assertEqual(o, p.path)
            self.assertEqual(u, str(p))


    def test_windows_urls(self):

        from rowgenerators.appurl.util import parse_url_to_dict, parse_file_to_uri
        from urllib.parse import  urlparse
        import pathlib
        import tabulate

        self.assertEqual(
            "file:///C:/Documents%20and%20Settings/davris/FileSchemeURIs.doc",
            str(parse_file_to_uri('C:\Documents and Settings\davris\FileSchemeURIs.doc'))
        )

        self.assertEqual(
            "file://laptop/My%20Documents/FileSchemeURIs.doc",
            str(parse_file_to_uri(r'\\laptop\My Documents\FileSchemeURIs.doc'))
        )

        rows = []

        urls = [
            'C:\Documents and Settings\davris\FileSchemeURIs.doc',
            'C:\\Documents and Settings\\davris\\FileSchemeURIs.doc',
            'c:/Documents and Settings/davris/FileSchemeURIs.doc',
            '\\\\laptop\\My Documents\\FileSchemeURIs.doc',
            r'\\laptop\My Documents\FileSchemeURIs.doc',
            '//laptop/My Documents/FileSchemeURIs.doc',
            'foo/bar/baz',
            '/foo/bar/baz',
            'http://a/b/c',
            'file:///a/b/s',
            'file://a/b/s',
            'file:/a/b/s',
            'file:a/b/s',
        ]

        for ou in urls:
            u = parse_app_url(ou)
            rows.append([ou,u, u.fspath])
            


        print(tabulate.tabulate(rows))

        return


        self.assertEqual('file:W:/metatab36/metatab-py/metatab/templates/metatab.csv',
                          str(parse_app_url(url)))

        url = 'N:/Desktop/metadata.csv#renter_cost'

        self.assertEqual('file:N:/Desktop/metadata.csv#renter_cost',
                          str(parse_app_url(url)))

    def test_query_urls(self):

        url='https://s3.amazonaws.com/private.library.civicknowledge.com/civicknowledge.com-rcfe_health-1/metadata.csv?AWSAccessKeyId=AKIAJFW23EPQCLXRU7DA&Signature=A39XhRP%2FTKAxv%2B%2F5vCubwWPDag0%3D&Expires=1494223447'

        u = Url(url)

        self.assertEqual('metadata.csv', str(u.resource_file))
        self.assertEqual('csv', str(u.resource_format))

        self.assertEqual('metadata.csv', str(u.target_file))
        self.assertEqual('csv', str(u.target_format))

    def test_socrata(self):

        u_s = 'https://data.lacounty.gov/api/views/8rdv-6nb6/rows.csv?accessType=DOWNLOAD'
        u = parse_app_url(u_s)
        self.assertIsInstance(parse_app_url(u_s), WebUrl)

        self.assertEqual('https://data.lacounty.gov/api/views/8rdv-6nb6/rows.csv?accessType=DOWNLOAD', str(u.resource_url))




    def test_targets(self):

        u=parse_app_url('http://library.metatab.org/example.com-example_data_package-2017-us-1.xlsx')

        tfu = u.join_target('random-names')

        self.assertEqual(
            'http://library.metatab.org/example.com-example_data_package-2017-us-1.xlsx#random-names',
            str(tfu))

        r = tfu.get_resource()

        self.assertTrue(str(r)
                        .endswith('library.metatab.org/example.com-example_data_package-2017-us-1.xlsx#random-names'))

        t = r.get_target()

        self.assertTrue(str(t)
                        .endswith('library.metatab.org/example.com-example_data_package-2017-us-1.xlsx#random-names'))


    def test_list(self):

        import rowgenerators.appurl

        u = parse_app_url(dirname(rowgenerators.appurl.__file__))

        self.assertTrue(len(list(u.list())) > 10)

        return

        u = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/test_data.zip')

        print(type(u))

        for su in u.list():
            for ssu in su.list():
                print(ssu)


    def test_stata(self):

        from itertools import islice, chain

        p = data_path('stata.dta')

        u = parse_app_url(p+'#categories=codes')

        print(type(u), u)

        print(u.get_target())

        print(type(u.generator))

        rows = list (chain(*islice(u.generator, 10)))

        self.assertEqual(140, len(rows))


    def test_x(self):
        from rowgenerators.appurl.url import UrlParts, Url
        from pprint import pprint

        us = 'https://user:pass@example2.com:8080/file.txt#segment&encoding=encoding&start=10'
        u = UrlParts(us)

        u2 = UrlParts(None, **u.dict)

        self.assertEqual(us, str(u2))

    def test_url_parts(self):
        from rowgenerators.appurl.url import UrlParts, Url
        from pprint import pprint

        records = [
            {
                'url_in':'https://user:pass@example1.com:8080/file.txt#segment&encoding=encoding&start=10',
                'attr': {
                    'proto': 'A',
                    'scheme': 'B',
                    'hostname': 'C',
                    'username': 'D',
                    'path': 'E.txt'
                },
                'url_out': 'A+B://D:pass@C:8080/E.txt#segment&encoding=encoding&start=10'
            },
            {
                'url_in': 'https://user:pass@example2.com:8080/file.txt#segment&encoding=encoding&start=10',
                'attr': {
                    'target_file': 'A',
                    'target_segment': 'B',
                    'encoding': 'C',
                    'start': 'D',
                    'end': 'E'
                },
                'url_out': 'https://user:pass@example2.com:8080/file.txt#A;B&encoding=C&end=E&start=D'
            },
            {
                'url_in': 'https://user:pass@example3.com:8080/file.txt#segment&encoding=encoding&start=10',
                'attr': {
                    'target_file': None,
                    'target_segment': None,
                    'encoding': None,
                    'start': None,
                    'end': None
                },
                'url_out': 'https://user:pass@example3.com:8080/file.txt'
            },
            {
                'url_in': 'https://example4.com/file.txt',
                'attr': {
                    'target_format': 'foobar',
                    'target_file': 'other_file.csv'

                },
                'url_out': 'https://example4.com/file.txt#other_file.csv&target_format=foobar'
            },
            {
                'url_in': 'https://example4.com/file.txt#tf;2',
                'url_out': 'https://example4.com/file.txt#tf;2'
            }
        ]

        # Test basic str and re-parse
        for d in records:
            for cls in (UrlParts, Url):
                u = cls(d['url_in'])
                self.assertEqual(d['url_in'], str(u))

        # Test changing URL through kwargs and attributes

        def u_attr(cls, d):
            u = cls(d['url_in'])
            for k, v in d.get('attr',{}).items():
                try:
                    setattr(u, k, v)
                except AttributeError:
                    raise AttributeError("Can't set {} on {} for {}".format(v, k, type(u)))
            return u

        def u_kw(cls, d):
            return cls(d['url_in'], **d.get('attr',{}))

        for d in records:
            for j, u in enumerate((u_attr(UrlParts, d), u_attr(Url, d), u_kw(UrlParts, d), u_kw(Url, d))):

                if d.get('url_out'):
                    try:
                        self.assertEqual(d['url_out'], str(u))
                    except Exception:
                        print('=========', j)
                        print(type(u))
                        print(d['url_in'])
                        pprint(u._parts)
                        print('=========')
                        raise
                else:
                    print(str(u))


#if __name__ == '__main__':
#    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from rowgenerators.source import Source

##
## FIXME This is probably broken
## But it's Socrata, so who really cares?

class SocrataSource(Source):
    """Iterates a CSV soruce from the JSON produced by Socrata  """

    def __init__(self, spec, dflo, cache, working_dir=None, env=None):

        super(SocrataSource, self).__init__(spec, cache)

        self._socrata_meta = None

        self._download_url = spec.url + '/rows.csv'

        self._csv_source = CsvSource(spec, dflo)

    @classmethod
    def download_url(cls, spec):
        return spec.url + '/rows.csv'

    @property
    def _meta(self):
        """Return the Socrata meta data, as a nested dict"""
        import requests

        if not self._socrata_meta:
            r = requests.get(self.spec.url)

            r.raise_for_status()

            self._socrata_meta = r.json()

        return self._socrata_meta

    @property
    def headers(self):
        """Return headers.  """

        return [c['fieldName'] for c in self._meta['columns']]

    datatype_map = {
        'text': 'str',
        'number': 'float',
    }

    @property
    def meta(self):
        """Return metadata """

        return {
            'title': self._meta.get('name'),
            'summary': self._meta.get('description'),
            'columns': [
                {
                    'name': c.get('fieldName'),
                    'position': c.get('position'),
                    'description': c.get('name') + '.' + c.get('description'),

                }
                for c in self._meta.get('columns')
                ]

        }

    def __iter__(self):

        self.start()

        for i, row in enumerate(self._csv_source):
            # if i == 0:
            #    yield self.headers

            yield row

        self.finish()




"""Pipes, pipe segments and piplines, for flowing data from sources to partitions.

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of the
Revised BSD License, included in this distribution as LICENSE.txt
"""

from .types import transform_generator

def test_exception(exception, bundle):
    print(exception)
    return None

@transform_generator
def test_tg(a,b,c, pipe, source, bundle):

    def _test_tg(v, row, **kwargs):

        print(v, a, b, c, row[0])
        return v

    return _test_tg


def test_transform_v(v):

    print(v)
    return v


def test_transform_vr(v, row):

    print(v, row[0])
    return v

def test_transform_vrh(v, row, header_d):

    print(v, row[0], header_d)
    return v


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from  rowgenerators.appurl import Url

class ApplicationUrl(Url):
    """Application URLs have weirdo schemes or protos"""

    reparse = False

    def __init__(self, url, **kwargs):
        raise NotImplementedError()
        super(ApplicationUrl, self).__init__(url, **kwargs)

    @classmethod
    def match(cls, url, **kwargs):
        return extract_proto(url) not in ['file', 'ftp', 'http', 'https']




# -*- coding: utf-8 -*-

"""
Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of the
Revised BSD License, included in this distribution as LICENSE.txt
"""

class RowGeneratorError(Exception):
    pass

class RowGeneratorConfigError(RowGeneratorError):
    """Row generator errors that result from the configuration of the rowgenerator,
    such as using a tab-based parser on a comma-based file"""

    def __init__(self, config_type: str, *args: object) -> None:
        super().__init__(*args)
        self.config_type = config_type

class SourceError(RowGeneratorError):
    pass

class DownloadError(SourceError):
    pass

class AccessError(DownloadError):
    """Got an acess error on download"""
    pass

class TextEncodingError(SourceError):
    pass

class SpecError(RowGeneratorError):
    pass


class ConfigurationError(RowGeneratorError):
    pass


class MissingCredentials(RowGeneratorError):
    pass

class SchemaError(RowGeneratorError):
    pass


class AppUrlError(Exception):
    pass



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
import inspect
from rowgenerators.source import Source


class IteratorSource(Source):
    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir)

        self.itr = ref

        if inspect.isfunction(self.itr):
            self.itr = self.itr()

    def __iter__(self):
        """ Iterate over all of the lines in the generator. """

        self.start()

        yield from self.itr

        self.finish()




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from rowgenerators.appurl.file.file import FileUrl
from rowgenerators.exceptions import AppUrlError


class ZipUrlError(AppUrlError):
    pass


class ZipUrl(FileUrl):
    """Zip URLS represent a zip file, as a local resource. """

    match_priority = FileUrl.match_priority - 10

    def __init__(self, url=None, downloader=None, **kwargs):

        super().__init__(url, downloader=downloader, **kwargs)

        if self.resource_format != 'zip':
            self.resource_format = 'zip'


    @property
    def target_file(self):
        """
        Returns the target file, which is usually stored in the first slot in the ``fragment``,
        but may have been overridden with a ``fragment_query``.

        :return:
        """

        if self._parts['target_file']:
            return self._parts['target_file']

        for ext in ('csv', 'xls', 'xlsx'):
            if self.resource_file.endswith('.' + ext + '.zip'):
                return self.resource_file.replace('.zip', '')

        # Want to return none, so get_files_from-zip can assume to use the first file in the archive.
        return None

    # Just a copy of the one from Url; looks like it must be reset because
    # the target_file prop was replaced
    @target_file.setter
    def target_file(self, v):
        self._parts['target_file'] = v

    def join_target(self, tf):
        """
        Joins the target ``tf`` by setting the value of the first slot of the fragment.

        :param tf:
        :return: a clone of this url with a new fragment.
        """
        u = self.clone()

        try:
            u.target_file = str(tf.path)
        except AttributeError:
            u.target_file = tf

        return u

    def get_resource(self):
        return self

    @property
    def zip_dir(self):
        """Directory that files will be extracted to"""

        from os.path import abspath

        cache_dir = self.downloader.cache.getsyspath('/')
        target_path = abspath(self.fspath)

        if target_path.startswith(cache_dir):  # Case when file is already in cache
            return str(self.fspath) + '_d'
        else:  # file is not in cache; it may exist elsewhere.
            return self.downloader.cache.getsyspath(target_path.lstrip('/'))+'_d'

    def get_target(self):
        """
        Extract the target file from the archive, store it in the cache, and return a file Url to the
        cached file.

        """

        from rowgenerators.appurl.url import parse_app_url
        from zipfile import ZipFile, BadZipFile
        from rowgenerators.appurl.util import copy_file_or_flo, ensure_dir
        from pathlib import Path
        from time import time


        assert self.zip_dir

        try:
            zf = ZipFile(str(self.fspath))
        except BadZipFile:
            raise ZipUrlError(f"Not a zip file: {str(self.fspath)} for url {str(self)}")

        self.target_file = ZipUrl.get_file_from_zip(self)

        target_path = Path(self.zip_dir).joinpath(self.target_file)
        ensure_dir(target_path.parent)

        def age(p):
            """File age in minutes"""
            try:
                return (time() - p.stat().st_mtime) / 60
            except:
                print("XXXX", p, p.exists())
                raise

        # Unpack the file if it dos not exist, or if the zile file is newer
        if not target_path.exists() or age(target_path) > age(Path(self.fspath)):
            with target_path.open('wb') as f, zf.open(self.target_file) as flo:
                copy_file_or_flo(flo, f)
            disp = 'copied'
        else:
            disp = 'extant'

        fq = self.frag_dict

        if 'resource_format' in fq:
            del fq['resource_format']

        if 'resource_file' in fq:
            del fq['resource_file']

        if 'target_file' in fq:
            del fq['target_file']

        tu =  parse_app_url(str(target_path),
                             scheme_extension=self.scheme_extension,
                             downloader=self.downloader,
                             **fq
                             )

        if self.target_format != tu.target_format:

            try:
                tu.target_format = self.target_format
            except AttributeError:
                pass # Some URLS don't allow resetting target type.

        tu._disp = disp

        return tu

    def list(self):
        """List the files in the referenced Zip file"""

        from zipfile import ZipFile


        if self.target_file:
            return list(self.set_target_segment(tl.target_segment) for tl in self.get_target().list())
        else:
            real_files = ZipUrl.real_files_in_zf(ZipFile(str(self.fspath)))
            return list(self.set_target_file(rf) for rf in real_files)

    @staticmethod
    def get_file_from_zip(url):
        """Given a file name that may be a regular expression, return the full name for the file
        from a zip archive"""

        from rowgenerators.exceptions import AppUrlError
        from zipfile import ZipFile, BadZipFile
        import re

        names = []
        try:
            zf = ZipFile(str(url.fspath))
        except BadZipFile:
            raise AppUrlError(f"Bad zip file: '{str(url.fspath)}' ")

        nl = list(ZipUrl.real_files_in_zf(zf))  # Old way, but maybe gets links? : list(zf.namelist())

        tf = url.target_file
        ts = url.target_segment

        if not nl:
            # sometimes real_files_in_zf doesn't work at all. I don't know why it does work,
            # so I certainly don't know why it does not.
            nl = list(zf.namelist())

        # the target_file may be a string, or a regular expression

        if tf:

            if tf.startswith('*'):
                # Common user error using a glob instead of a regex
                tf = tf.replace('*', '.*')

            try:
                names = list([e for e in nl if re.search(tf, e)
                          and not (e.startswith('__') or e.startswith('.'))
                          ])
            except Exception as e:
                raise

            if len(names) > 0:
                return names[0]


        # The segment, if it exists, can only be an integer, and should probably be
        # '0' to indicate the first file. This clause is probably a bad idea, since
        # andy other integer is probably meaningless.
        if ts:
            try:
                return nl[int(ts)]

            except (IndexError, ValueError):
                pass

        # Just return the first file in the archive.
        if not tf and not ts:
            return nl[0]
        else:

            raise ZipUrlError("Could not find file in Zip {} for target='{}' nor segment='{}'"
                              .format(url.fspath, url.target_file, url.target_segment))

    @staticmethod
    def real_files_in_zf(zf):
        """Return a list of internal paths of real files in a zip file, based on the 'external_attr' values"""
        from os.path import basename

        for e in zf.infolist():



            # Get rid of __MACOS and .DS_whatever
            if basename(e.filename).startswith('__') or basename(e.filename).startswith('.'):
                continue

            # I really don't understand external_attr, but no one else seems to either,
            # so we're just hacking here.
            # e.external_attr>>31&1 works when the archive has external attrs set, and a dir heirarchy
            # e.external_attr==0 works in cases where there are no external attrs set
            # e.external_attr==32 is true for some single-file archives.
            # e.external_attr==128 has been true at least once ...
            if bool(e.external_attr >> 31 & 1 or e.external_attr == 0 or e.external_attr == 32 or e.external_attr == 128):
                yield e.filename


    @classmethod
    def _match(cls, url, **kwargs):

        return url.resource_format == 'zip' or kwargs.get('force_archive')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

from setuptools import find_packages, setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.5',
    'Topic :: Software Development :: Debuggers',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


setup(
    name='rowgenerators',
    version="0.9.19",
    description='Generate row data from a variety of file formats',
    long_description=readme,
    packages=find_packages(),
    install_requires=[
        'pyproj',
        'python-dateutil',
        'boto',
        'codegen',
        'decorator',
        'filelock',
        'fs >= 2',
        'geoid',
        'livestats',
        'meta',
        'petl',
        'requests',
        'tabulate',
        'xlrd<2',
        'aniso8601',
        'geopandas',
        'pyyaml',
        'h5py',
        'wrapt'
    ],
    extras_require={
        'geo': ['fiona', 'shapely','pyproj', 'pyproject']
    },
    test_requires=['aniso8601', 'dateutil', 'fiona', 'shapely','pyproj', 'pyproject', 'contexttimer'],
    entry_points={
        'console_scripts': [
            'rowgen=rowgenerators.cli:rowgen',
            'rowgen-generators=rowgenerators.cli:listrowgen',
            'rowgen-urls=rowgenerators.appurl.cli:appurl',
            'rowgen-valuestypes=rowgenerators.valuetype.cli:valuetypes'
        ],


        'appurl.urls': [
            "* = rowgenerators.appurl.url:Url",
            # Web Urls
            "http: = rowgenerators.appurl.web.web:WebUrl",
            "https: = rowgenerators.appurl.web.web:WebUrl",
            "ftp: = rowgenerators.appurl.web.web:FtpUrl",
            "s3: = rowgenerators.appurl.web.s3:S3Url",
            "socrata+ = rowgenerators.appurl.web.socrata:SocrataUrl",
            #
            # Archive Urls
            ".zip = rowgenerators.appurl.archive.zip:ZipUrl",
            #
            # File Urls
            ".csv = rowgenerators.appurl.file.csv:CsvFileUrl",
            ".xlsx = rowgenerators.appurl.file.excel:ExcelFileUrl",
            ".xls = rowgenerators.appurl.file.excel:ExcelFileUrl",
            ".dta = rowgenerators.appurl.file.stata:StataUrl",
            "file: = rowgenerators.appurl.file.file:FileUrl",
            "program+ = rowgenerators.appurl.file.program:ProgramUrl",
            "python: = rowgenerators.appurl.file.python:PythonUrl",
            ".py = rowgenerators.appurl.file.python:PythonUrl",

            "shape+ = rowgenerators.appurl.file.shapefile:ShapefileUrl",
            "/docs.google.com/spreadsheets/ = rowgenerators.appurl.web.google:GoogleSpreadsheetUrl",
            "gs: = rowgenerators.appurl.web.google:GoogleSpreadsheetUrl",
            "gs+ = rowgenerators.appurl.web.google:GoogleSpreadsheetUrl",

            #Sql Alchemy
            "oracle: = rowgenerators.appurl.sql:OracleSql",
            "sql: = rowgenerators.appurl.sql:InlineSqlUrl",
            ".h5 = rowgenerators.appurl.file.hdf5:Hdf5Url",
            ".hdf5 = rowgenerators.appurl.file.hdf5:Hdf5Url",
        ],

        'rowgenerators': [
            "<iterator> = rowgenerators.generator.iterator:IteratorSource",
            "<generator> = rowgenerators.generator.generator:GeneratorSource",
            ".csv = rowgenerators.generator.csv:CsvSource",
            ".tsv = rowgenerators.generator.delimited:TsvSource",
            ".pipe = rowgenerators.generator.delimited:PipeSource",
            ".xlsx = rowgenerators.generator.excel:ExcelSource",
            ".xls =  rowgenerators.generator.excel:ExcelSource",
            ".dta =  rowgenerators.generator.stata:StataSource",
            "program+ = rowgenerators.generator.program:ProgramSource",
            "shape+ = rowgenerators.generator.shapefile:ShapefileSource",
            ".geojson = rowgenerators.generator.shapefile:GeoJsonSource",
            ".shp = rowgenerators.generator.shapefile:ShapefileSource",
            "<PythonUrl> = rowgenerators.generator.python:PythonSource",
            "fixed+ = rowgenerators.generator.fixed:FixedSource",
            "<Sql> = rowgenerators.generator.sql:SqlSource",
            ".h5 = rowgenerators.generator.hdf5:Hdf5Source",
            ".hdf5 = rowgenerators.generator.hdf5:Hdf5Source",
        ],
    },

    author="Eric Busboom",
    author_email='eric@civicknowledge.com',
    url='https://github.com/Metatab/rowgenerator.git',
    license='MIT',
    classifiers=classifiers
)



from __future__ import print_function

import unittest



from rowgenerators.appurl.url import parse_app_url
from rowgenerators.appurl.util import parse_url_to_dict, parse_file_to_uri

from rowgenerators.appurl.web.download import Downloader



class TestUrlParse(unittest.TestCase):
    def test_fragment(self):
        def u(frag):
            return parse_app_url("http://example.com/foo.csv#" + frag)

        self.assertEqual({}, u('a').fragment_query)
        self.assertEqual({}, u('a;b').fragment_query)
        self.assertEqual({'foo': 'bar'}, u('a;b&foo=bar').fragment_query)
        self.assertEqual({'foo': 'bar'}, u('a;&foo=bar').fragment_query)
        self.assertEqual({'foo': 'bar'}, u('a&foo=bar').fragment_query)
        self.assertEqual({'foo': 'bar'}, u('&foo=bar').fragment_query)
        self.assertEqual({'foo': 'bar'}, u(';&foo=bar').fragment_query)

        url = u('a;b&encoding=utf8')

        self.assertEqual('utf8', url.encoding)

        url.encoding = 'ascii'
        url.start = 5

        self.assertEqual('http://example.com/foo.csv#a;b&encoding=ascii&start=5', str(url))

        url = u('a;b&target_format=foo&resource_format=bar')

        self.assertEqual('foo', url.target_format)
        self.assertEqual('bar', url.resource_format)

        us = 'http://public.source.civicknowledge.com/example.com/sources/test_data.zip#unicode-latin1.csv&encoding=latin1'

        url = parse_app_url(us)

        self.assertEqual('latin1', url.encoding)
        self.assertEqual('latin1', url.get_resource().encoding)
        ru =  url.get_resource()
        tu = ru.get_target()
        self.assertEqual('latin1',tu.encoding)

    def test_parse_file_fragment(self):

        fn = '/example.com/foo.csv#&a=b'

        u = parse_app_url(fn)

        print(u.fragment_query)

        d = parse_url_to_dict('file://'+fn)
        print(d['path'])
        print(d['fragment_query'])

        print('----')
        print(parse_file_to_uri(fn))



    def test_fragment_2(self):
        url = parse_app_url(
            'http://public.source.civicknowledge.com/example.com/sources/renter_cost_excel07.zip#target_format=xlsx')

        self.assertEqual('zip', url.resource_format)
        self.assertEqual('renter_cost_excel07.zip', url.target_file)
        self.assertEqual('xlsx', url.target_format)

        r = url.get_resource()

        self.assertEqual('xlsx', r.target_format)
        self.assertEqual('zip', r.resource_format)

    def test_two_extensions(self):
        u_s = 'http://public.source.civicknowledge.com/example.com/sources/simple-example.csv.zip'

        u = parse_app_url(u_s, Downloader.get_instance())

        self.assertEqual('simple-example.csv.zip', u.resource_file)
        self.assertEqual('simple-example.csv.zip', u.target_file)
        self.assertEqual('zip', u.target_format)

        r = u.get_resource()
        print(type(r), r)
        self.assertEqual('simple-example.csv.zip', r.resource_file)
        self.assertEqual('simple-example.csv', r.target_file)

        self.assertEqual('csv', r.target_format)

    def test_shapefile_url(self):
        from rowgenerators.appurl.web.web import WebUrl

        u_s = 'shapefile+http://public.source.civicknowledge.com.s3.amazonaws.com/example.com/geo/Parks_SD.zip'

        u = parse_app_url(u_s)

        self.assertIsInstance(u, WebUrl, type(u))

        r = u.get_resource()

        self.assertTrue(str(r).startswith('shapefile+'))

    def test_s3_url(self):
        from rowgenerators.appurl.web.s3 import S3Url

        url_str = 's3://bucket/a/b/c/file.csv'

        u = parse_app_url(url_str)

        self.assertIsInstance(u, S3Url, type(u))

    def test_python_url(self):
        from rowgenerators.appurl.file import python
        from rowgenerators import get_generator
        from types import ModuleType

        import sys

        foo = ModuleType('foo')
        sys.modules['foo'] = foo
        foo.bar = ModuleType('bar')
        sys.modules['foo.bar'] = foo.bar
        foo.bar.baz = ModuleType('baz')
        sys.modules['foo.bar.baz'] = foo.bar.baz

        def foobar(*args, **kwargs):
            for i in range(10):
                yield i

        foo.bar.baz.foobar = foobar

        u = parse_app_url("python:foo.bar.baz#foobar")

        g = get_generator(u)

        self.assertEqual(45, sum(list(g)))

    def x_test_sql_url(self):

        url = """oracle://coredev:{CREATIONS_PW}@creations.bc.edu:1521/SISCNV#"FINAID"."COMMENTS_TEXT_LINE_RECORD" """

        u = parse_app_url(url)

        g = u.generator

        print(type(g), g)

        dsn_url = "oracle://coredev:{CREATIONS_PW}@creations.bc.edu:1521/SISCNV"

        url = 'sql://creations#sql://creations#SELECT * FROM  "FINAID"."COMMENTS_TEXT_LINE_RECORD"'

        dsn_dict = { 'creations' : dsn_url}

        u = parse_app_url(url, dsns=dsn_dict).get_resource().get_target()

        print(type(u), u)


    def test_interpolate(self):

        from rowgenerators.appurl.web.download import Downloader

        Downloader.context = {
            'PASSWORD' : 'correcthorse',
            'HOST': 'sajak'
        }

        u = parse_app_url('http://user:{PASSWORD}@{HOST}/foo/bar')

        print(u.interpolate())

    def test_maintain_target(self):

        b = 'https://www2.census.gov/programs-surveys/acs/summary_file/2016/data/5_year_seq_by_state/RhodeIsland/Tracts_Block_Groups_Only'

        us = b+'/20165ri0001000.zip#m20165ri0001000.txt&target_format=csv'


        u = parse_app_url(us)

        print (u)


    def test_google_url(self):
        from rowgenerators.appurl.web.google import GoogleSpreadsheetUrl

        us1 = 'https://drive.google.com/open?id=1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w'
        us2 = 'gs+https://drive.google.com/open?id=1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w'
        us3 = 'https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing'
        us4 = 'https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing'
        us5 = 'https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit#gid=801701031'
        us6 = 'gs://1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w#801701031'

        for us in  [us1, us2, us3, us4, us5, us6]:
            u = GoogleSpreadsheetUrl(us)

            self.assertEqual('1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w', u.key)

        for us in  [us1, us2, us3, us4]:
            u = GoogleSpreadsheetUrl(us)

            self.assertIsNone(u.gid)

        for us in  [us5, us6]:
            u = GoogleSpreadsheetUrl(us)

            self.assertEqual('801701031', u.gid)


        # parse_app_url

        us1 = 'gs+https://drive.google.com/open?id=1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w'
        us2 = 'gs+https://drive.google.com/open?id=1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w'
        us3 = 'gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing'
        us4 = 'gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit?usp=sharing'
        us5 = 'gs+https://docs.google.com/spreadsheets/d/1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w/edit#gid=801701031'
        us6 = 'gs://1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w#801701031'

        for us in  [us1, us2, us3, us4, us5, us6]:
            u = parse_app_url(us)
            self.assertEqual('1VGEkgXXmpWya7KLkrAPHp3BLGbXibxHqZvfn9zA800w', u.key)

        for us in  [us1, us2, us3, us4]:
            u = parse_app_url(us)
            self.assertIsNone(u.gid)

        for us in  [us5, us6]:
            u = parse_app_url(us)
            self.assertEqual('801701031', u.gid)

    def test_no_netloc(self):
        from rowgenerators import Url

        url_strs = ['proto://a/b/c', 'proto:/a/b/c', 'proto:a/b/c']

        #for us in url_strs:
        #    u = Url(us)
        #    print("{} | {}".format(u.netloc, u.path))

        print('---')

        url_strs = ['proto://a/b/c', 'proto:/a/b/c', 'proto:a/b/c']

        for us in url_strs:
            u = Url(us, no_netloc=True, no_abs_path=True)
            self.assertEqual(u.netloc,'')
            self.assertEqual(u.path,'a/b/c')

            u = Url(us, no_netloc=True).remove_netloc().absify_path()
            self.assertEqual(u.path, '/a/b/c')





if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""


"""

from .url import Url, parse_app_url, match_url_classes
from .util import get_cache



"""


Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

from .core import *

class MeasureMixin(object):
    pass


class IntMeasure(IntValue, MeasureMixin):
    role = ROLE.MEASURE
    vt_code = 'measure/int'

class LongMeasure(LongValue, MeasureMixin):
    role = ROLE.MEASURE
    vt_code = 'mmeasure/long'

class FloatMeasure(FloatValue, MeasureMixin):
    role = ROLE.MEASURE
    vt_code = 'measure'


class ZScoreVT(FloatMeasure):
    """A Z score"""
    desc = 'Z Score'
    vt_code = 'zscore'

class ArealDensityVT(FloatMeasure):
    """A general areal density"""
    desc = 'Density'
    vt_code = 'density'


class CountVT(IntMeasure):
    vt_code = 'count'
    desc = 'Count'
    def __init__(self,v):
        pass

class RatioVT(FloatMeasure):
    """A general ratio, with values that may exceed 1"""
    vt_code = 'ratio'
    desc = 'Ratio'

class ProportionVT(FloatMeasure):
    """A general ratio of two other values, from 0 to 1"""
    vt_code = 'proportion'
    desc = 'Proportion'

    def __new__(cls, v):

        o = FloatMeasure.__new__(cls, v)

        if bool(o) and float(o) > 1:
            return FailedValue(v, ValueError("Proportion values must be less than 1"))

        return o

    @property
    def rate(self):
        return self

    @property
    def percent(self):
        return PercentageVT(self*100)

class RateVT(ProportionVT):
    """A general ratio of two other values, from 0 to 1"""
    vt_code = 'rate'
    desc = 'Rate, 0->1'

class PercentageVT(FloatMeasure):
    """Percentage, expressed as 0 to 100. """
    vt_code = 'pct'
    desc = 'Percentage 0->100'

    def __new__(cls, v):

        if isinstance(v, str) and '%' in v:
            v = v.strip('%')

        return FloatValue.__new__(cls, v)

    def init(self):
        pass

    @property
    def rate(self):
        return ProportionVT(self / 100.0)

    @property
    def percent(self):
        return self



measure_value_types = {
    "measure": FloatMeasure,
    "measure/float": FloatMeasure,
    "measure/int": IntMeasure,
    "measure/long": LongMeasure,
    "density": ArealDensityVT,
    "count": CountVT,
    "ratio": RatioVT,
    "rate": RateVT,
    "proportion": ProportionVT,
    "pct": PercentageVT,
    "percent": PercentageVT,
    "z": ZScoreVT,
}